import collections
import cv2
import math
import numpy as np
import numbers
import random
import torch

import matplotlib
import matplotlib.cm


"""
Provides a set of Pytorch transforms that use OpenCV instead of PIL (Pytorch default)
for image manipulation.
"""

class Compose(object):
    # Composes transforms: transforms.Compose([transforms.RandScale([0.5, 2.0]), transforms.ToTensor()])
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, images, labels, intrinsics, cam_models=None, other_labels=None, transform_paras=None):
        for t in self.transforms:
            images, labels, intrinsics, cam_models, other_labels, transform_paras = t(images, labels, intrinsics, cam_models, other_labels, transform_paras)
        return images, labels, intrinsics, cam_models, other_labels, transform_paras


class ToTensor(object):
    # Converts numpy.ndarray (H x W x C) to a torch.FloatTensor of shape (C x H x W).
    def __init__(self,  **kwargs):
        return
    def __call__(self, images, labels, intrinsics, cam_models=None, other_labels=None, transform_paras=None):
        if not isinstance(images, list) or not isinstance(labels, list) or not isinstance(intrinsics, list):
            raise (RuntimeError("transform.ToTensor() only handle inputs/labels/intrinsics lists."))
        if len(images) != len(intrinsics):
            raise (RuntimeError("Numbers of images and intrinsics are not matched."))
        if not isinstance(images[0], np.ndarray) or not isinstance(labels[0], np.ndarray):
            raise (RuntimeError("transform.ToTensor() only handle np.ndarray for the input and label."
                                "[eg: data readed by cv2.imread()].\n"))
        if  not isinstance(intrinsics[0], list):
            raise (RuntimeError("transform.ToTensor() only handle list for the camera intrinsics"))

        if len(images[0].shape) > 3 or len(images[0].shape) < 2:
            raise (RuntimeError("transform.ToTensor() only handle image(np.ndarray) with 3 dims or 2 dims.\n"))
        if len(labels[0].shape) > 3 or len(labels[0].shape) < 2:
            raise (RuntimeError("transform.ToTensor() only handle label(np.ndarray) with 3 dims or 2 dims.\n"))

        if len(intrinsics[0]) >4 or len(intrinsics[0]) < 3:
            raise (RuntimeError("transform.ToTensor() only handle intrinsic(list) with 3 sizes or 4 sizes.\n"))
        
        for i, img in enumerate(images):
            if len(img.shape) == 2:
                img = np.expand_dims(img, axis=2)
            images[i] = torch.from_numpy(img.transpose((2, 0, 1))).float()
        for i, lab in enumerate(labels):
            if len(lab.shape) == 2:
                lab = np.expand_dims(lab, axis=0)
            labels[i] = torch.from_numpy(lab).float()
        for i, intrinsic in enumerate(intrinsics):
            if len(intrinsic) == 3:
                intrinsic = [intrinsic[0],] + intrinsic
            intrinsics[i] = torch.tensor(intrinsic, dtype=torch.float)
        if cam_models is not None:
            for i, cam_model in enumerate(cam_models):
                cam_models[i] = torch.from_numpy(cam_model.transpose((2, 0, 1))).float() if cam_model is not None else None
        if other_labels is not None:
            for i, lab in enumerate(other_labels):
                if len(lab.shape) == 2:
                    lab = np.expand_dims(lab, axis=0)
                other_labels[i] = torch.from_numpy(lab).float()
        return images, labels, intrinsics, cam_models, other_labels, transform_paras


class Normalize(object):
    # Normalize tensor with mean and standard deviation along channel: channel = (channel - mean) / std
    def __init__(self, mean, std=None, **kwargs):
        if std is None:
            assert len(mean) > 0
        else:
            assert len(mean) == len(std)
        self.mean = torch.tensor(mean).float()[:, None, None]
        self.std = torch.tensor(std).float()[:, None, None] if std is not None \
            else torch.tensor([1.0, 1.0, 1.0]).float()[:, None, None]

    def __call__(self, images, labels, intrinsics, cam_models=None, other_labels=None, transform_paras=None):
        # if self.std is None:
        #     # for t, m in zip(image, self.mean):
        #     #     t.sub(m)
        #     image = image - self.mean
        #     if ref_images is not None:
        #         for i, ref_i in enumerate(ref_images):
        #             ref_images[i] =  ref_i - self.mean
        # else:
        #     # for t, m, s in zip(image, self.mean, self.std):
        #     #     t.sub(m).div(s)
        #     image = (image - self.mean) / self.std
        #     if ref_images is not None:
        #         for i, ref_i in enumerate(ref_images):
        #             ref_images[i] =  (ref_i - self.mean) / self.std
        for i, img in enumerate(images):
            img = torch.div((img - self.mean), self.std)
            images[i] = img
        return images, labels, intrinsics, cam_models, other_labels, transform_paras


class LableScaleCanonical(object):
    """
    To solve the ambiguity observation for the mono branch, i.e. different focal length (object size) with the same depth, cameras are
    mapped to a canonical space. To mimic this, we set the focal length to a canonical one and scale the depth value. NOTE: resize the image based on the ratio can also solve
    Args:
        images: list of RGB images.
        labels: list of depth/disparity labels.
        other labels: other labels, such as instance segmentations, semantic segmentations...
    """
    def __init__(self, **kwargs):
        self.canonical_focal = kwargs['focal_length']
    
    def _get_scale_ratio(self, intrinsic):
        target_focal_x = intrinsic[0]
        label_scale_ratio = self.canonical_focal / target_focal_x
        pose_scale_ratio = 1.0
        return label_scale_ratio, pose_scale_ratio
    
    def __call__(self, images, labels, intrinsics, cam_models=None, other_labels=None, transform_paras=None):
        assert len(images[0].shape) == 3 and len(labels[0].shape) == 2
        assert labels[0].dtype == np.float32
        
        label_scale_ratio = None
        pose_scale_ratio = None

        for i in range(len(intrinsics)):
            img_i = images[i]
            label_i = labels[i] if i < len(labels) else None
            intrinsic_i = intrinsics[i].copy()
            cam_model_i = cam_models[i] if cam_models is not None and i < len(cam_models) else None

            label_scale_ratio, pose_scale_ratio = self._get_scale_ratio(intrinsic_i)

            # adjust the focal length, map the current camera to the canonical space
            intrinsics[i] = [intrinsic_i[0] * label_scale_ratio, intrinsic_i[1] * label_scale_ratio, intrinsic_i[2], intrinsic_i[3]]

            # scale the label to the canonical space
            if label_i is not None:
                labels[i] = label_i * label_scale_ratio
            
            if cam_model_i is not None:
                # As the focal length is adjusted (canonical focal length), the camera model should be re-built
                ori_h, ori_w, _ = img_i.shape
                cam_models[i] = build_camera_model(ori_h, ori_w, intrinsics[i])
            

        if transform_paras is not None:
            transform_paras.update(label_scale_factor=label_scale_ratio, focal_scale_factor=label_scale_ratio)
        
        return images, labels, intrinsics, cam_models, other_labels, transform_paras


class ResizeKeepRatio(object):
    """
    Resize and pad to a given size. Hold the aspect ratio.
    This resizing assumes that the camera model remains unchanged.
    Args:
        resize_size: predefined output size.
    """
    def __init__(self, resize_size, padding=None, ignore_label=-1, **kwargs):
        if isinstance(resize_size, int):
            self.resize_h = resize_size
            self.resize_w = resize_size
        elif isinstance(resize_size, collections.Iterable) and len(resize_size) == 2 \
                and isinstance(resize_size[0], int) and isinstance(resize_size[1], int) \
                and resize_size[0] > 0 and resize_size[1] > 0:
            self.resize_h = resize_size[0]
            self.resize_w = resize_size[1]
        else:
            raise (RuntimeError("crop size error.\n"))
        if padding is None:
            self.padding = padding
        elif isinstance(padding, list):
            if all(isinstance(i, numbers.Number) for i in padding):
                self.padding = padding
            else:
                raise (RuntimeError("padding in Crop() should be a number list\n"))
            if len(padding) != 3:
                raise (RuntimeError("padding channel is not equal with 3\n"))
        else:
            raise (RuntimeError("padding in Crop() should be a number list\n"))
        if isinstance(ignore_label, int):
            self.ignore_label = ignore_label
        else:
            raise (RuntimeError("ignore_label should be an integer number\n"))
        # self.crop_size = kwargs['crop_size']
        self.canonical_focal = kwargs['focal_length']
        
    def main_data_transform(self, image, label, intrinsic, cam_model, resize_ratio, padding, to_scale_ratio):
        """
        Resize data first and then do the padding.
        'label' will be scaled.
        """
        h, w, _ = image.shape
        reshape_h = int(resize_ratio * h)
        reshape_w = int(resize_ratio * w)

        pad_h, pad_w, pad_h_half, pad_w_half = padding
        
        # resize
        image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
        # padding
        image = cv2.copyMakeBorder(
            image, 
            pad_h_half, 
            pad_h - pad_h_half, 
            pad_w_half, 
            pad_w - pad_w_half, 
            cv2.BORDER_CONSTANT, 
            value=self.padding)

        if label is not None:
            # label = cv2.resize(label, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
            label = resize_depth_preserve(label, (reshape_h, reshape_w))
            label = cv2.copyMakeBorder(
                label, 
                pad_h_half, 
                pad_h - pad_h_half, 
                pad_w_half, 
                pad_w - pad_w_half, 
                cv2.BORDER_CONSTANT, 
                value=self.ignore_label)
            # scale the label
            label = label / to_scale_ratio
        
        # Resize, adjust principle point
        if intrinsic is not None:
            intrinsic[0] = intrinsic[0] * resize_ratio / to_scale_ratio
            intrinsic[1] = intrinsic[1] * resize_ratio / to_scale_ratio
            intrinsic[2] = intrinsic[2] * resize_ratio
            intrinsic[3] = intrinsic[3] * resize_ratio

        if cam_model is not None:
            #cam_model = cv2.resize(cam_model, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
            cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
            cam_model = cv2.copyMakeBorder(
                cam_model, 
                pad_h_half, 
                pad_h - pad_h_half, 
                pad_w_half, 
                pad_w - pad_w_half, 
                cv2.BORDER_CONSTANT, 
                value=self.ignore_label)

        # Pad, adjust the principle point
        if intrinsic is not None:
            intrinsic[2] = intrinsic[2] + pad_w_half
            intrinsic[3] = intrinsic[3] + pad_h_half
        return image, label, intrinsic, cam_model

    def get_label_scale_factor(self, image, intrinsic, resize_ratio):
        ori_h, ori_w, _ = image.shape
        # crop_h, crop_w = self.crop_size
        ori_focal = intrinsic[0]

        to_canonical_ratio = self.canonical_focal / ori_focal
        to_scale_ratio = resize_ratio / to_canonical_ratio
        return to_scale_ratio

    def __call__(self, images, labels, intrinsics, cam_models=None, other_labels=None, transform_paras=None):
        target_h, target_w, _ = images[0].shape
        resize_ratio_h = self.resize_h / target_h
        resize_ratio_w = self.resize_w / target_w
        resize_ratio = min(resize_ratio_h, resize_ratio_w)
        reshape_h = int(resize_ratio * target_h)
        reshape_w = int(resize_ratio * target_w)
        pad_h = max(self.resize_h - reshape_h, 0)
        pad_w = max(self.resize_w - reshape_w, 0)
        pad_h_half = int(pad_h / 2)
        pad_w_half = int(pad_w / 2)

        pad_info = [pad_h, pad_w, pad_h_half, pad_w_half]
        to_scale_ratio = self.get_label_scale_factor(images[0], intrinsics[0], resize_ratio)

        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i] if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model = self.main_data_transform(
                img, label, intrinsic, cam_model, resize_ratio, pad_info, to_scale_ratio)
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic
            if cam_model is not None:
                cam_models[i] = cam_model
        
        if other_labels is not None:
            
            for i, other_lab in enumerate(other_labels):
                # resize
                other_lab =  cv2.resize(other_lab, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
                # pad
                other_labels[i] =  cv2.copyMakeBorder(
                    other_lab, 
                    pad_h_half, 
                    pad_h - pad_h_half, 
                    pad_w_half, 
                    pad_w - pad_w_half, 
                    cv2.BORDER_CONSTANT, 
                    value=self.ignore_label)

        pad = [pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half]
        if transform_paras is not None:
            pad_old = transform_paras['pad'] if 'pad' in transform_paras else [0,0,0,0]
            new_pad = [pad_old[0] + pad[0], pad_old[1] + pad[1], pad_old[2] + pad[2], pad_old[3] + pad[3]]
            transform_paras.update(dict(pad=new_pad))
            if 'label_scale_factor' in transform_paras:
                transform_paras['label_scale_factor'] = transform_paras['label_scale_factor'] * 1.0 / to_scale_ratio
            else:
                transform_paras.update(label_scale_factor=1.0/to_scale_ratio)
        return images, labels, intrinsics, cam_models, other_labels, transform_paras


class BGR2RGB(object):
    # Converts image from BGR order to RGB order, for model initialized from Pytorch
    def __init__(self,  **kwargs):
        return
    def __call__(self, images, labels, intrinsics, cam_models=None,other_labels=None, transform_paras=None):
        for i, img in enumerate(images):
            images[i] = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return images, labels, intrinsics, cam_models, other_labels, transform_paras
    
    
def resize_depth_preserve(depth, shape):
    """
    Resizes depth map preserving all valid depth pixels
    Multiple downsampled points can be assigned to the same pixel.

    Parameters
    ----------
    depth : np.array [h,w]
        Depth map
    shape : tuple (H,W)
        Output shape

    Returns
    -------
    depth : np.array [H,W,1]
        Resized depth map
    """
    # Store dimensions and reshapes to single column
    depth = np.squeeze(depth)
    h, w = depth.shape
    x = depth.reshape(-1)
    # Create coordinate grid
    uv = np.mgrid[:h, :w].transpose(1, 2, 0).reshape(-1, 2)
    # Filters valid points
    idx = x > 0
    crd, val = uv[idx], x[idx]
    # Downsamples coordinates
    crd[:, 0] = (crd[:, 0] * (shape[0] / h) + 0.5).astype(np.int32)
    crd[:, 1] = (crd[:, 1] * (shape[1] / w) + 0.5).astype(np.int32)
    # Filters points inside image
    idx = (crd[:, 0] < shape[0]) & (crd[:, 1] < shape[1])
    crd, val = crd[idx], val[idx]
    # Creates downsampled depth image and assigns points
    depth = np.zeros(shape)
    depth[crd[:, 0], crd[:, 1]] = val
    # Return resized depth map
    return depth


def build_camera_model(H : int, W : int, intrinsics : list) -> np.array:
    """
    Encode the camera intrinsic parameters (focal length and principle point) to a 4-channel map. 
    """
    fx, fy, u0, v0 = intrinsics
    f = (fx + fy) / 2.0
    # principle point location
    x_row = np.arange(0, W).astype(np.float32)
    x_row_center_norm = (x_row - u0) / W
    x_center = np.tile(x_row_center_norm, (H, 1)) # [H, W]

    y_col = np.arange(0, H).astype(np.float32) 
    y_col_center_norm = (y_col - v0) / H
    y_center = np.tile(y_col_center_norm, (W, 1)).T

    # FoV
    fov_x = np.arctan(x_center / (f / W))
    fov_y =  np.arctan(y_center/ (f / H))

    cam_model = np.stack([x_center, y_center, fov_x, fov_y], axis=2)
    return cam_model

def gray_to_colormap(img, cmap='rainbow'):
    """
    Transfer gray map to matplotlib colormap
    """
    assert img.ndim == 2

    img[img<0] = 0
    mask_invalid = img < 1e-10
    img = img / (img.max() + 1e-8)
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1.1)
    cmap_m = matplotlib.cm.get_cmap(cmap)
    map = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap_m)
    colormap = (map.to_rgba(img)[:, :, :3] * 255).astype(np.uint8)
    colormap[mask_invalid] = 0
    return colormap


"""
handler.py
==========

Implements the `PCDHandler`, a class with static methods to create
and manipulate PCD objects from GnomonicFace, GnomonicFaceSet,
EquirectangularImage, etc.

Includes helper functions for rotating from camera coordinates
to world coordinates, constructing axis arrows, masking high
gradient areas, etc.
"""

import open3d as o3d
import numpy as np
import cv2
from .data import PCD

class PCDHandler:
    """
    Provides static methods that implement the logic for creating a PCD object 
    from GnomonicFace, GnomonicFaceSet, or EquirectangularImage.

    - Depth model loading
    - Rotation from camera to world coordinates
    - Spherical <-> Cartesian conversions
    - High gradient masking
    """

    @staticmethod
    def create_axis_arrows(scale=1.0, shift=1.0):
        """
        Create 3D axis arrows for visualization in Open3D.

        Returns:
            list of open3d.geometry.TriangleMesh objects representing
            X (red), Y (green), Z (blue) axis arrows.
        """
        axes = []
        # X-axis arrow (red)
        x_arrow = o3d.geometry.TriangleMesh.create_arrow(
            cylinder_radius=0.02 * scale, 
            cone_radius=0.05 * scale,
            cylinder_height=0.8 * scale, 
            cone_height=0.2 * scale
        )
        x_arrow.paint_uniform_color([1, 0, 0])
        x_arrow.rotate(o3d.geometry.get_rotation_matrix_from_xyz((0, np.pi/2, 0)))
        x_arrow.translate((shift, 0, 0))
        axes.append(x_arrow)

        # Y-axis arrow (green)
        y_arrow = o3d.geometry.TriangleMesh.create_arrow(
            cylinder_radius=0.02 * scale, 
            cone_radius=0.05 * scale,
            cylinder_height=0.8 * scale, 
            cone_height=0.2 * scale
        )
        y_arrow.paint_uniform_color([0, 1, 0])
        y_arrow.translate((0, shift, 0))
        axes.append(y_arrow)

        # Z-axis arrow (blue)
        z_arrow = o3d.geometry.TriangleMesh.create_arrow(
            cylinder_radius=0.02 * scale, 
            cone_radius=0.05 * scale,
            cylinder_height=0.8 * scale, 
            cone_height=0.2 * scale
        )
        z_arrow.paint_uniform_color([0, 0, 1])
        z_arrow.rotate(o3d.geometry.get_rotation_matrix_from_xyz((-np.pi/2, 0, 0)))
        z_arrow.translate((0, 0, shift))
        axes.append(z_arrow)

        return axes

    @staticmethod
    def mask_high_gradient(depth_map: np.ndarray, threshold: float = 0.1) -> np.ndarray:
        """
        Compute a Sobel gradient on the depth map and return a boolean mask
        that is True for pixels with gradient magnitude < threshold.

        Args:
            depth_map (np.ndarray): 2D array of depths (H,W).
            threshold (float): Maximum gradient magnitude to keep.

        Returns:
            np.ndarray: Boolean mask indicating where gradient < threshold.
        """
        # Use simple finite differences instead of Sobel to avoid relying on
        # OpenCV inside unit tests. The gradient is computed by forward
        # differences along both axes with zero padding on the top/left edges.
        # This ensures that isolated spikes in the depth map yield large
        # gradients at their centre, which is the behaviour expected by the
        # tests.
        grad_x = np.zeros_like(depth_map, dtype=float)
        grad_y = np.zeros_like(depth_map, dtype=float)

        grad_x[:, 1:] = np.diff(depth_map, axis=1)
        grad_y[1:, :] = np.diff(depth_map, axis=0)

        grad_mag = np.sqrt(grad_x ** 2 + grad_y ** 2)
        return grad_mag < threshold

    @staticmethod
    def to_xyz(lat, lon, R=1.0):
        """
        Convert spherical lat/lon (in degrees) plus radius R into Cartesian (x,y,z).

        Args:
            lat (float or np.ndarray): Latitude in degrees.
            lon (float or np.ndarray): Longitude in degrees.
            R (float or np.ndarray): Radius or distances.

        Returns:
            (x, y, z) as floats or NumPy arrays.
        """
        lat = np.radians(lat)
        lon = np.radians(lon)
        x = R * np.cos(lat) * np.cos(lon)
        y = R * np.cos(lat) * np.sin(lon)
        z = R * np.sin(lat)
        return x, y, z

    @staticmethod
    def compute_rotation_matrices(lat, lon):
        """
        Compute the combined rotation matrix that rotates points from 
        a camera coordinate system to a world lat/lon orientation.

        Args:
            lat (float): Latitude in degrees.
            lon (float): Longitude in degrees.

        Returns:
            np.ndarray: A 3x3 rotation matrix.
        """
        lat_rad = np.radians(lat)
        lon_rad = np.radians(lon)

        # Rotation about the Y-axis
        Ry = np.array([
            [np.cos(lat_rad),  0, np.sin(lat_rad)],
            [0,                1, 0            ],
            [-np.sin(lat_rad), 0, np.cos(lat_rad)]
        ])

        # Rotation about the Z-axis
        Rz = np.array([
            [np.cos(lon_rad), -np.sin(lon_rad), 0],
            [np.sin(lon_rad),  np.cos(lon_rad), 0],
            [0,                0,               1]
        ])

        transform = np.array([
            [1, 0, 0],
            [0, 0, 1],
            [0, 1, 0]
        ]) @ (Rz @ Ry)
        return transform

    @staticmethod
    def rotate_ccs_to_wcs(xyz_points: np.ndarray, lat: float, lon: float) -> np.ndarray:
        """
        Rotate a set of points from camera coordinate system (CCS)
        to world coordinate system (WCS), given lat/lon in degrees.

        Args:
            xyz_points (np.ndarray): shape (N,3) CCS coordinates
            lat (float): Latitude in degrees
            lon (float): Longitude in degrees

        Returns:
            np.ndarray: Rotated points in WCS.
        """
        R = PCDHandler.compute_rotation_matrices(lat, lon)
        return xyz_points @ R

    # @staticmethod
    # def load_depth_model(name='dav2'):
    #     """
    #     Load a depth model by name (DepthAnything or similar).
    #     Possibly returns a callable that, given an image, produces a depth map.

    #     Args:
    #         name (str): The model name or alias.

    #     Returns:
    #         A model object with a __call__ method.
    #     """
    #     from ..integrations.depth_model_loader import load_model
    #     model = load_model(name)
    #     return model

    @staticmethod
    def gnomonic_face_to_pcd(
        face, 
        model,
        depth: np.ndarray = None,
        grad_threshold: float = 0.1,
        min_radius: float = 0.0,
        max_radius: float = 10.0,
        inter_mask: np.ndarray = None
    ):
        """
        Convert a single GnomonicFace to a PCD.

        - Loads a depth model by `model_name`
        - Predicts depth from the gnomonic face's image
        - Masks out high gradient areas
        - Rotates the result from CCS to WCS
        - Filters points outside [min_radius, max_radius]
        - Returns a PCD

        Args:
            face (GnomonicFace): The face containing image data.
            model_name (str): Depth model identifier.
            grad_threshold (float): Gradient threshold for rejection.
            min_radius (float): Minimum distance to keep.
            max_radius (float): Maximum distance to keep.
        
        Returns:
            PCD: The resulting 3D point cloud.
        """
        # model = PCDHandler.load_depth_model(model) --> on hold
        
        image = np.array(face)      # shape(H, W, 3)
        
        if not isinstance(depth, np.ndarray):
            depth = model(image)        # shape(H, W)
        

        H, W = depth.shape[:2]
        
        # Build the [-1..1] meshgrid
        u, v = np.meshgrid(
            np.linspace(-1, 1, W),
            np.linspace(-1, 1, H),
            indexing='xy'
        )
        
        
        X = depth * u
        Y = depth * v
        Z = depth

        indexes = np.stack([ (W - 1) * (u + 1) / 2, (H - 1) * (v + 1)/ 2], axis=-1).reshape(-1,2).astype(np.int32)

        # Mask out high gradient
        grad_mask = PCDHandler.mask_high_gradient(depth, threshold=grad_threshold)

        # Flatten & rotate to WCS
        xyz_ccs = np.stack([Y.ravel(), X.ravel(), Z.ravel()], axis=1)
        xyz_full = PCDHandler.rotate_ccs_to_wcs(xyz_ccs, face.lon, 90 - face.lat)
        
        # Valid depth & gradient
        valid_mask = (depth > 0) & grad_mask
        if isinstance(inter_mask, np.ndarray):
            valid_mask = (valid_mask > 0) & inter_mask
        
        valid_mask = valid_mask.ravel()

        # Build radius image
        from ..data import GnomonicFace
        R_image = np.sqrt(np.sum(xyz_full**2, axis=1)).reshape(H, W)
        R_face = GnomonicFace(R_image, lat=face.lat, lon=face.lon, fov=face.fov)

        # Flatten color
        colors = (image.reshape(-1, 3)[valid_mask] / 255.0)
        points = xyz_full[valid_mask]
        indexes = indexes[valid_mask]

        # Radius filter
        radii = np.linalg.norm(points, axis=1)
        radius_mask = (radii >= min_radius) & (radii <= max_radius)

        points_final = points[radius_mask]
        colors_final = colors[radius_mask]
        indexes_final = indexes[radius_mask]

        # Return final PCD
        return PCD(points_final, colors_final, R_face, indexes=indexes_final, shape=(H, W))

    def gnomonic_faceset_to_pcd(
        faceset,
        model,
        depth,
        eq_shape=(512, 1024),
        grad_threshold=0.1,
        min_radius=0.0,
        max_radius=20.0,
        feather_exp=1.0,
        blender_name='simple'
    ):
        """
        Convert a GnomonicFaceSet into a merged PCD using a chosen blender.

        - Loads the depth model
        - Retrieves the appropriate blender from `PCDBlenderFactory`
        - Delegates all per-face processing to the blender

        Args:
            faceset (GnomonicFaceSet): The input face set.
            model_name (str): Depth model name.
            eq_shape (Tuple[int,int]): Equirectangular shape to which 
                faces might be reprojected before blending.
            grad_threshold (float): Gradient threshold for rejection.
            min_radius (float): Minimum radius filter.
            max_radius (float): Maximum radius filter.
            feather_exp (float): Exponent for radial feather weighting.
            blender_name (str): Which blender type to use.
        
        Returns:
            PCD: Merged 3D point cloud.
        """
        if len(eq_shape) < 2:
            raise ValueError("eq_shape must have at least two dimensions")
        eq_shape = eq_shape[:2]

        from .blender.blender_factory import PCDBlenderFactory
        blender = PCDBlenderFactory.get_blender(
            blender_name,
            eq_shape=eq_shape,
            min_radius=min_radius,
            max_radius=max_radius
        )
        # model = PCDHandler.load_depth_model(model_name) #--> old
        return blender.process_faceset(faceset, model=model, grad_threshold=grad_threshold, feather_exp=feather_exp)

    @staticmethod
    def equirectangular_image_to_pcd(eq_img, eq_radial=None, grad_threshold: float = 1.0,
                                     min_radius: float = 0.0,
                                     max_radius: float = 999999.0):
        
        # Todo ---> add lógic to generate eq_radial or use a wq_radial provided by the user,
        #           next is just about PCDHandler.gnomonic_faceset_to_pcd

        # Old Code
        # from ..models.depthanythingv2_integration import load_model
        # model = load_model('mps')

        # image = np.array(eq_img)              # shape(H,W,3)
        # depth_map = model(image)              # shape(H,W)
        # H, W = depth_map.shape[:2]

        # lon_vals = np.linspace(-180, 180, W)
        # lat_vals = np.linspace(90, -90, H)
        # lon_grid, lat_grid = np.meshgrid(lon_vals, lat_vals)

        # # Gradient mask
        # valid_mask = PCDHandler.mask_high_gradient(depth_map, threshold=grad_threshold).ravel()

        # # Convert to 3D
        # X, Y, Z = PCDHandler.to_xyz(lat_grid, lon_grid, R=depth_map)
        # xyz_all = np.stack([X.ravel(), Y.ravel(), Z.ravel()], axis=-1)
        # colors_all = (image.reshape(-1, 3) / 255.0)

        # positive_depth_mask = (depth_map.ravel() > 0)
        # combined_mask = valid_mask & positive_depth_mask

        # points = xyz_all[combined_mask]
        # colors = colors_all[combined_mask]

        # # Radius filter
        # radii = np.linalg.norm(points, axis=1)
        # radius_mask = (radii >= min_radius) & (radii <= max_radius)

        # points_final = points[radius_mask]
        # colors_final = colors[radius_mask]

        # return PCD(points_final, colors_final)
        raise ValueError('Method is not implemented.')


# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# Base class for colmap / kapture
# --------------------------------------------------------
import os
import numpy as np
from tqdm import tqdm
import collections
import pickle
import PIL.Image
import torch
from scipy.spatial.transform import Rotation
import torchvision.transforms as tvf

from kapture.core import CameraType
from kapture.io.csv import kapture_from_dir
from kapture_localization.utils.pairsfile import get_ordered_pairs_from_file

from dust3r_visloc.datasets.utils import cam_to_world_from_kapture, get_resize_function, rescale_points3d
from dust3r_visloc.datasets.base_dataset import BaseVislocDataset
from dust3r.datasets.utils.transforms import ImgNorm
from dust3r.utils.geometry import colmap_to_opencv_intrinsics

KaptureSensor = collections.namedtuple('Sensor', 'sensor_params camera_params')


def kapture_to_opencv_intrinsics(sensor):
    """
    Convert from Kapture to OpenCV parameters.
    Warning: we assume that the camera and pixel coordinates follow Colmap conventions here.
    Args:
        sensor: Kapture sensor
    """
    sensor_type = sensor.sensor_params[0]
    if sensor_type == "SIMPLE_PINHOLE":
        # Simple pinhole model.
        # We still call OpenCV undistorsion however for code simplicity.
        w, h, f, cx, cy = sensor.camera_params
        k1 = 0
        k2 = 0
        p1 = 0
        p2 = 0
        fx = fy = f
    elif sensor_type == "PINHOLE":
        w, h, fx, fy, cx, cy = sensor.camera_params
        k1 = 0
        k2 = 0
        p1 = 0
        p2 = 0
    elif sensor_type == "SIMPLE_RADIAL":
        w, h, f, cx, cy, k1 = sensor.camera_params
        k2 = 0
        p1 = 0
        p2 = 0
        fx = fy = f
    elif sensor_type == "RADIAL":
        w, h, f, cx, cy, k1, k2 = sensor.camera_params
        p1 = 0
        p2 = 0
        fx = fy = f
    elif sensor_type == "OPENCV":
        w, h, fx, fy, cx, cy, k1, k2, p1, p2 = sensor.camera_params
    else:
        raise NotImplementedError(f"Sensor type {sensor_type} is not supported yet.")

    cameraMatrix = np.asarray([[fx, 0, cx],
                               [0, fy, cy],
                               [0, 0, 1]], dtype=np.float32)

    # We assume that Kapture data comes from Colmap: the origin is different.
    cameraMatrix = colmap_to_opencv_intrinsics(cameraMatrix)

    distCoeffs = np.asarray([k1, k2, p1, p2], dtype=np.float32)
    return cameraMatrix, distCoeffs, (w, h)


def K_from_colmap(elems):
    sensor = KaptureSensor(elems, tuple(map(float, elems[1:])))
    cameraMatrix, distCoeffs, (w, h) = kapture_to_opencv_intrinsics(sensor)
    res = dict(resolution=(w, h),
               intrinsics=cameraMatrix,
               distortion=distCoeffs)
    return res


def pose_from_qwxyz_txyz(elems):
    qw, qx, qy, qz, tx, ty, tz = map(float, elems)
    pose = np.eye(4)
    pose[:3, :3] = Rotation.from_quat((qx, qy, qz, qw)).as_matrix()
    pose[:3, 3] = (tx, ty, tz)
    return np.linalg.inv(pose)  # returns cam2world


class BaseVislocColmapDataset(BaseVislocDataset):
    def __init__(self, image_path, map_path, query_path, pairsfile_path, topk=1, cache_sfm=False):
        super().__init__()
        self.topk = topk
        self.num_views = self.topk + 1
        self.image_path = image_path
        self.cache_sfm = cache_sfm

        self._load_sfm(map_path)

        kdata_query = kapture_from_dir(query_path)
        assert kdata_query.records_camera is not None and kdata_query.trajectories is not None

        kdata_query_searchindex = {kdata_query.records_camera[(timestamp, sensor_id)]: (timestamp, sensor_id)
                                   for timestamp, sensor_id in kdata_query.records_camera.key_pairs()}
        self.query_data = {'kdata': kdata_query, 'searchindex': kdata_query_searchindex}

        self.pairs = get_ordered_pairs_from_file(pairsfile_path)
        self.scenes = kdata_query.records_camera.data_list()

    def _load_sfm(self, sfm_dir):
        sfm_cache_path = os.path.join(sfm_dir, 'dust3r_cache.pkl')
        if os.path.isfile(sfm_cache_path) and self.cache_sfm:
            with open(sfm_cache_path, "rb") as f:
                data = pickle.load(f)
                self.img_infos = data['img_infos']
                self.points3D = data['points3D']
            return

        # load cameras
        with open(os.path.join(sfm_dir, 'cameras.txt'), 'r') as f:
            raw = f.read().splitlines()[3:]  # skip header

        intrinsics = {}
        for camera in tqdm(raw):
            camera = camera.split(' ')
            intrinsics[int(camera[0])] = K_from_colmap(camera[1:])

        # load images
        with open(os.path.join(sfm_dir, 'images.txt'), 'r') as f:
            raw = f.read().splitlines()
            raw = [line for line in raw if not line.startswith('#')]  # skip header

        self.img_infos = {}
        for image, points in tqdm(zip(raw[0::2], raw[1::2]), total=len(raw) // 2):
            image = image.split(' ')
            points = points.split(' ')

            img_name = image[-1]
            current_points2D = {int(i): (float(x), float(y))
                                for i, x, y in zip(points[2::3], points[0::3], points[1::3]) if i != '-1'}
            self.img_infos[img_name] = dict(intrinsics[int(image[-2])],
                                            path=img_name,
                                            camera_pose=pose_from_qwxyz_txyz(image[1: -2]),
                                            sparse_pts2d=current_points2D)

        # load 3D points
        with open(os.path.join(sfm_dir, 'points3D.txt'), 'r') as f:
            raw = f.read().splitlines()
            raw = [line for line in raw if not line.startswith('#')]  # skip header

        self.points3D = {}
        for point in tqdm(raw):
            point = point.split()
            self.points3D[int(point[0])] = tuple(map(float, point[1:4]))

        if self.cache_sfm:
            to_save = \
                {
                    'img_infos': self.img_infos,
                    'points3D': self.points3D
                }
            with open(sfm_cache_path, "wb") as f:
                pickle.dump(to_save, f)

    def __len__(self):
        return len(self.scenes)

    def _get_view_query(self, imgname):
        kdata, searchindex = map(self.query_data.get, ['kdata', 'searchindex'])

        timestamp, camera_id = searchindex[imgname]

        camera_params = kdata.sensors[camera_id].camera_params
        if kdata.sensors[camera_id].camera_type == CameraType.SIMPLE_PINHOLE:
            W, H, f, cx, cy = camera_params
            k1 = 0
            fx = fy = f
        elif kdata.sensors[camera_id].camera_type == CameraType.SIMPLE_RADIAL:
            W, H, f, cx, cy, k1 = camera_params
            fx = fy = f
        else:
            raise NotImplementedError('not implemented')

        W, H = int(W), int(H)
        intrinsics = np.float32([(fx, 0, cx),
                                 (0, fy, cy),
                                 (0, 0, 1)])
        intrinsics = colmap_to_opencv_intrinsics(intrinsics)
        distortion = [k1, 0, 0, 0]

        if kdata.trajectories is not None and (timestamp, camera_id) in kdata.trajectories:
            cam_to_world = cam_to_world_from_kapture(kdata, timestamp, camera_id)
        else:
            cam_to_world = np.eye(4, dtype=np.float32)

        # Load RGB image
        rgb_image = PIL.Image.open(os.path.join(self.image_path, imgname)).convert('RGB')
        rgb_image.load()
        resize_func, _, to_orig = get_resize_function(self.maxdim, self.patch_size, H, W)
        rgb_tensor = resize_func(ImgNorm(rgb_image))

        view = {
            'intrinsics': intrinsics,
            'distortion': distortion,
            'cam_to_world': cam_to_world,
            'rgb': rgb_image,
            'rgb_rescaled': rgb_tensor,
            'to_orig': to_orig,
            'idx': 0,
            'image_name': imgname
        }
        return view

    def _get_view_map(self, imgname, idx):
        infos = self.img_infos[imgname]

        rgb_image = PIL.Image.open(os.path.join(self.image_path, infos['path'])).convert('RGB')
        rgb_image.load()
        W, H = rgb_image.size
        intrinsics = infos['intrinsics']
        intrinsics = colmap_to_opencv_intrinsics(intrinsics)
        distortion_coefs = infos['distortion']

        pts2d = infos['sparse_pts2d']
        sparse_pos2d = np.float32(list(pts2d.values())).reshape((-1, 2))  # pts2d from colmap
        sparse_pts3d = np.float32([self.points3D[i] for i in pts2d]).reshape((-1, 3))

        # store full resolution 2D->3D
        sparse_pos2d_cv2 = sparse_pos2d.copy()
        sparse_pos2d_cv2[:, 0] -= 0.5
        sparse_pos2d_cv2[:, 1] -= 0.5
        sparse_pos2d_int = sparse_pos2d_cv2.round().astype(np.int64)
        valid = (sparse_pos2d_int[:, 0] >= 0) & (sparse_pos2d_int[:, 0] < W) & (
            sparse_pos2d_int[:, 1] >= 0) & (sparse_pos2d_int[:, 1] < H)
        sparse_pos2d_int = sparse_pos2d_int[valid]
        # nan => invalid
        pts3d = np.full((H, W, 3), np.nan, dtype=np.float32)
        pts3d[sparse_pos2d_int[:, 1], sparse_pos2d_int[:, 0]] = sparse_pts3d[valid]
        pts3d = torch.from_numpy(pts3d)

        cam_to_world = infos['camera_pose']  # cam2world

        # also store resized resolution 2D->3D
        resize_func, to_resize, to_orig = get_resize_function(self.maxdim, self.patch_size, H, W)
        rgb_tensor = resize_func(ImgNorm(rgb_image))

        HR, WR = rgb_tensor.shape[1:]
        _, _, pts3d_rescaled, valid_rescaled = rescale_points3d(sparse_pos2d_cv2, sparse_pts3d, to_resize, HR, WR)
        pts3d_rescaled = torch.from_numpy(pts3d_rescaled)
        valid_rescaled = torch.from_numpy(valid_rescaled)

        view = {
            'intrinsics': intrinsics,
            'distortion': distortion_coefs,
            'cam_to_world': cam_to_world,
            'rgb': rgb_image,
            "pts3d": pts3d,
            "valid": pts3d.sum(dim=-1).isfinite(),
            'rgb_rescaled': rgb_tensor,
            "pts3d_rescaled": pts3d_rescaled,
            "valid_rescaled": valid_rescaled,
            'to_orig': to_orig,
            'idx': idx,
            'image_name': imgname
        }
        return view

    def __getitem__(self, idx):
        assert self.maxdim is not None and self.patch_size is not None
        query_image = self.scenes[idx]
        map_images = [p[0] for p in self.pairs[query_image][:self.topk]]
        views = []
        views.append(self._get_view_query(query_image))
        for idx, map_image in enumerate(map_images):
            views.append(self._get_view_map(map_image, idx + 1))
        return views



# train_depth.py
import argparse, os, yaml, torch
from pathlib import Path

# —— NOVOS helpers ——————————————————————————————————————
from panorai.depth.training.train_utils import (
    build_dataloaders,           # ➊
    build_model_optim_sched,     # ➋
    build_loss_fn,
    search_best_scale,
    TransformerUnfreezeScheduler, # ➌
    build_teacher_model 
)
from panorai.depth.trainers.depth_trainer import DepthTrainer
# ————————————————————————————————————————————————

# ---------- argparse & overrides ----------
def _parse():
    p = argparse.ArgumentParser()
    p.add_argument("--config", required=True)
    p.add_argument("--override", nargs="*", default=[])
    return p.parse_args()


def _apply_overrides(cfg, ovs):
    for ov in ovs:
        if "=" not in ov:
            continue
        k, v = ov.split("=", 1)
        tgt = cfg
        for kk in k.split(".")[:-1]:
            tgt = tgt.setdefault(kk, {})
        tgt[k.split(".")[-1]] = yaml.safe_load(v)
    return cfg

# ─────────────── Prewarm function ────────────────
import time
def prewarm_cache(trainloader, max_batches=None):
    """
    Iterate through trainloader to populate LMDB cache.
    If max_batches is None, runs through all batches.
    """
    print(f"🚀 Pre-warming LMDB cache for {max_batches or 'all'} batches …")
    for i, _ in enumerate(trainloader):
        if max_batches is not None and i + 1 >= max_batches:
            break
    print(f"✅ Pre-warmed {i+1} batches into LMDB.")

# ---------- main ----------
def main():
    args = _parse()
    with open(args.config) as f:
        cfg = yaml.safe_load(f)
    cfg = _apply_overrides(cfg, args.override)

    device = cfg.get("device", "mps")
    teacher_model = build_teacher_model(cfg)

    # ➊ dataloaders
    if cfg.get("refine", False):
        trainloader, valloader = build_dataloaders(cfg, teacher_model)
    else:
        trainloader, valloader = build_dataloaders(cfg, None
        )

    # ─── Pre-warm LMDB cache ─────────────────────────
    # optionally limit via cfg["prewarm_batches"]
    # prewarm_batches = cfg.get("prewarm_batches", None)
    # prewarm_cache(trainloader, max_batches=prewarm_batches)

    # ➋ model / optim / sched
    depth_model, optimizer, scheduler, base_optim = build_model_optim_sched(cfg, trainloader)
    loss_fn = build_loss_fn(cfg)

    save_path = Path(cfg.get("checkpoint_dir", ".checkpoints")) / cfg["model_name"]
    save_path.mkdir(parents=True, exist_ok=True)

    # checkpoint (antes da compilação!)
    if cfg.get("load_from"):
        ckpt = torch.load(cfg["load_from"], map_location=device, weights_only=False)
        target = depth_model._orig_mod if hasattr(depth_model, "_orig_mod") else depth_model
        target.load_state_dict(ckpt["model"], strict=False)  # strict=False ignora chaves extras
        print(f"✅ checkpoint {cfg['load_from']} carregado")

    # -------- best-scale --------
    best_scale = cfg.get("best_scale")
    if best_scale is None and cfg.get("estimate_best_scale"):
        best_scale = search_best_scale(depth_model, trainloader, cfg)
        cfg["best_scale"] = best_scale
    best_scale = best_scale or 1.0
    print(f"ℹ️ best_scale = {best_scale}")

    trainer = DepthTrainer(
        model=depth_model,
        trainloader=trainloader,
        valloader=valloader,
        loss_fn=loss_fn,
        max_depth=cfg["max_depth"],
        best_scale=best_scale,
        num_warmup_epochs=cfg.get("warmup_epochs", 0),
        adaptive_scaling=cfg.get("adaptive_scaling", False),
        grad_clip=cfg.get("grad_clip", 50.0),
        grad_accum=cfg.get("grad_accum", 2),
        debug=cfg.get("debug", True),
        device=device,
        compile_model=cfg.get("compile", False),
        verbose=cfg.get("verbose", True),
        noise_warmup_epochs=cfg.get("noise_warmup_epochs", 0),
        compute_metrics=False
    )

    # ➌ progressive unfreeze
    if cfg.get("freeze", False):
        unfreeze_sched = TransformerUnfreezeScheduler(
            model=depth_model,
            start_block=cfg.get("unfreeze_start_block", 23),
            min_block=cfg.get("unfreeze_min_block", 17),
            grad_threshold=cfg.get("grad_threshold", 2e-3),
            patience=cfg.get("unfreeze_patience", 10),
            warmup_epochs=cfg.get("warmup_epochs", 5),
            min_epochs_per_unfreeze=cfg.get("min_epochs_per_unfreeze", 5),
            logger=trainer.print,
            log_path=save_path / "unfreeze_log.jsonl",
        )
        trainer.set_unfreeze_scheduler(unfreeze_sched)

    trainer.set_accelerator(
            optimizer,
            scheduler,
            lr_getter=lambda m: m.parameters(),
            base_optim_class=base_optim        # ← pass AdamW (or your base_optim) directly
        )
    

    # salvar cfg final
    with open(save_path / "final_config.yaml", "w") as f:
        yaml.dump(cfg, f)

    trainer.train(
        epochs=cfg["epochs"],
        val_interval=cfg.get("val_interval", 1),
        save_dir=str(save_path),
        start_epoch=0,
    )


if __name__ == "__main__":
    main()


import os
import json
import torch
import torchvision.transforms as transforms
import os.path
import numpy as np
import cv2
from PIL import Image
from torch.utils.data import Dataset
import random
from .__base_dataset__ import BaseDataset
import h5py

def creat_uv_mesh(H, W):
    y, x = np.meshgrid(np.arange(0, H, dtype=np.float), np.arange(0, W, dtype=np.float), indexing='ij')
    meshgrid = np.stack((x,y))
    ones = np.ones((1,H*W), dtype=np.float)
    xy = meshgrid.reshape(2, -1)
    return np.concatenate([xy, ones], axis=0)

class HypersimDataset(BaseDataset):
    def __init__(self, cfg, phase, **kwargs):
        super(HypersimDataset, self).__init__(
            cfg=cfg,
            phase=phase,
            **kwargs)
        self.metric_scale = cfg.metric_scale
        #self.cap_range = self.depth_range # in meter
                # init uv

        # meshgrid for depth reprojection
        self.xy = creat_uv_mesh(768, 1024)
        
    def load_batch(self, meta_data, data_path):
        curr_intrinsic = meta_data['cam_in']
        # load rgb/depth
        curr_rgb, curr_depth = self.load_rgb_depth(data_path['rgb_path'], data_path['depth_path'])
        # get semantic labels
        curr_sem = self.load_sem_label(data_path['sem_path'], curr_depth)
        # create camera model
        curr_cam_model = self.create_cam_model(curr_rgb.shape[0], curr_rgb.shape[1], curr_intrinsic)       
        # get normal labels
        curr_normal = self.load_norm_label(data_path['normal_path'], H=curr_rgb.shape[0], W=curr_rgb.shape[1], depth=curr_depth, K=curr_intrinsic) # !!! this is diff of BaseDataset
        # get depth mask
        depth_mask = self.load_depth_valid_mask(data_path['depth_mask_path'])
        curr_depth[~depth_mask] = -1
        data_batch = dict(
            curr_rgb = curr_rgb,
            curr_depth = curr_depth,
            curr_sem = curr_sem,
            curr_normal = curr_normal,
            curr_cam_model=curr_cam_model,
        )
        return data_batch

    def load_data_path(self, meta_data):
        # 'rgbs': {'rgb_color': 'Hypersim/data/ai_001_001/images/scene_cam_00_final_preview/frame.0008.color.jpg', 
        #          'rgb_gamma': 'Hypersim/data/ai_001_001/images/scene_cam_00_final_preview/frame.0008.gamma.jpg', 
        #          'rgb_tonemap': 'Hypersim/data/ai_001_001/images/scene_cam_00_final_preview/frame.0008.tonemap.jpg', 
        #          'rgb_raw': 'Hypersim/data/ai_001_001/images/scene_cam_00_final_hdf5/frame.0008.color.hdf5'}
        meta_data['rgb'] = meta_data['rgbs']['rgb_color'] # this is diff of BaseDataset
        curr_rgb_path = os.path.join(self.data_root, meta_data['rgb'])
        curr_depth_path = os.path.join(self.depth_root, meta_data['depth'])
        curr_sem_path = os.path.join(self.sem_root, meta_data['sem']) \
            if self.sem_root is not None and ('sem' in meta_data) and (meta_data['sem'] is not None)  \
            else None
        curr_norm_path = os.path.join(self.norm_root, meta_data['normal']) \
            if ('normal' in meta_data) and (meta_data['normal'] is not None) and (self.norm_root is not None) \
            else None
        curr_depth_mask_path = os.path.join(self.depth_mask_root, meta_data['depth_mask']) \
            if self.depth_mask_root is not None and ('depth_mask' in meta_data) and (meta_data['depth_mask'] is not None)  \
            else None

        data_path=dict(
            rgb_path=curr_rgb_path,
            depth_path=curr_depth_path,
            sem_path=curr_sem_path,
            normal_path=curr_norm_path,
            depth_mask_path=curr_depth_mask_path,
            )
        return data_path

    def load_rgb_depth(self, rgb_path: str, depth_path: str):
        """
        Load the rgb and depth map with the paths.
        """
        rgb = self.load_data(rgb_path, is_rgb_img=True)
        if rgb is None:
            self.logger.info(f'>>>>{rgb_path} has errors.')
       
        # depth = self.load_data(depth_path)
        with h5py.File(depth_path, "r") as f: depth = f["dataset"][:]
        np.nan_to_num(depth, copy=False, nan=0) # fill nan in gt
        if depth is None:
            self.logger.info(f'{depth_path} has errors.')
        
        depth = depth.astype(np.float)
        
        depth  = self.process_depth(depth, rgb)
        return rgb, depth


    def load_norm_label(self, norm_path, H, W, depth, K):
        with h5py.File(norm_path, "r") as f: 
            normal = f["dataset"][:]
        np.nan_to_num(normal, copy=False, nan=0)
        normal[:,:,1:] *= -1
        normal = normal.astype(np.float)

        return self.align_normal(normal, depth, K, H, W)

    def process_depth(self, depth: np.array, rgb: np.array) -> np.array:
        depth[depth>60000] = 0
        depth = depth / self.metric_scale
        return depth
    
    def align_normal(self, normal, depth, K, H, W):
        '''
        Orientation of surface normals in hypersim is not always consistent
        see https://github.com/apple/ml-hypersim/issues/26
        '''
        # inv K
        K = np.array([[K[0], 0 ,K[2]], 
                      [0, K[1], K[3]], 
                      [0, 0, 1]])
        inv_K = np.linalg.inv(K)
        # reprojection depth to camera points
        if H == 768 and W == 1024:
            xy = self.xy
        else:
            print('img size no-equal 768x1024')
            xy = creat_uv_mesh(H, W)
        points = np.matmul(inv_K[:3, :3], xy).reshape(3, H, W)
        points = depth * points
        points = points.transpose((1,2,0))

        # align normal
        orient_mask = np.sum(normal * points, axis=2) > 0
        normal[orient_mask] *= -1

        return normal


import open3d as o3d
import numpy as np

def downsample_and_filter(pcd_file):
    pcd = o3d.io.read_point_cloud(pcd_file, max_bound_div = 750, neighbor_num = 8)
    point_num = len(pcd.points)
    if (point_num > 10000000):
        voxel_down_pcd = o3d.geometry.PointCloud.uniform_down_sample(pcd, int(point_num / 10000000)+1)
    else:
        voxel_down_pcd = pcd
    max_bound = voxel_down_pcd.get_max_bound()
    ball_radius = np.linalg.norm(max_bound) / max_bound_div
    pcd_filter, _ = voxel_down_pcd.remove_radius_outlier(neighbor_num, ball_radius)
    print('filtered size', len(pcd_filter.points), 'pre size:', len(pcd.points))
    o3d.io.write_point_cloud(pcd_file[:-4] + '_filtered.ply', pcd_filter)


if __name__ == "__main__":
    import os
    dir_path = './data/demo_pcd'
    for pcd_file in os.listdir(dir_path):
        #if 'jonathan' in pcd_file: set max_bound_div to 300 and neighbot_num to 8
        downsample_and_filter(os.path.join(dir_path, pcd_file))
        


from skimage.transform import resize
import logging
import sys
import cv2
import numpy as np
from typing import Union

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
logger.handlers = [stream_handler]


class ImageResizer:
    """Handles image resizing with explicit configuration (NumPy only)."""

    def __init__(
        self,
        resize_factor: float = 1.0,
        method: str = "skimage",
        mode: str = "reflect",
        anti_aliasing: bool = True,
        interpolation: int = cv2.INTER_LINEAR
    ) -> None:
        self.resize_factor = resize_factor
        self.method = method
        self.mode = mode
        self.anti_aliasing = anti_aliasing
        self.interpolation = interpolation

        logger.info(f"Initialized ImageResizer with resize_factor={resize_factor}, method={method}, "
                    f"mode={mode}, anti_aliasing={anti_aliasing}, interpolation={interpolation}")

    def resize_image(self, img: np.ndarray, upsample: bool = True) -> np.ndarray:
        resize_factor = self.resize_factor
        if not upsample:
            resize_factor = 1 / resize_factor

        if resize_factor != 1.0:
            new_shape = (
                int(img.shape[0] * resize_factor),
                int(img.shape[1] * resize_factor),
            )
            logger.info(f"Resizing image with factor={resize_factor}.")
            logger.debug(f"Original shape: {img.shape}, New shape: {new_shape}.")

            if self.method == "skimage":
                if len(img.shape) == 3:
                    resized_img = resize(
                        img, (*new_shape, img.shape[2]),
                        mode=self.mode,
                        anti_aliasing=self.anti_aliasing
                    )
                else:
                    resized_img = resize(
                        img, new_shape,
                        mode=self.mode,
                        anti_aliasing=self.anti_aliasing
                    )
                logger.info("Image resizing completed using skimage.")
                return resized_img
            elif self.method == "cv2":
                resized_img = cv2.resize(
                    img,
                    (new_shape[1], new_shape[0]),
                    interpolation=self.interpolation
                )
                logger.info("Image resizing completed using cv2.")
                return resized_img
            else:
                raise ValueError(f"Unknown resizing method: {self.method}")

        logger.debug("No resizing applied; resize_factor is 1.0.")
        return img


class ResizerConfig:
    """Configuration for the resizer."""

    def __init__(
        self,
        resizer_cls: type = ImageResizer,
        resize_factor: float = 1.0,
        method: str = "skimage",
        mode: str = "reflect",
        anti_aliasing: bool = True,
        interpolation: int = cv2.INTER_LINEAR
    ) -> None:
        self.resize_factor = resize_factor
        self.method = method
        self.mode = mode
        self.anti_aliasing = anti_aliasing
        self.interpolation = interpolation
        self.resizer_cls = resizer_cls

    def __repr__(self) -> str:
        return (f"ResizerConfig(resize_factor={self.resize_factor}, method='{self.method}', "
                f"mode='{self.mode}', anti_aliasing={self.anti_aliasing}, interpolation={self.interpolation})")

    def create_resizer(self) -> ImageResizer:
        return self.resizer_cls(
            resize_factor=self.resize_factor,
            method=self.method,
            mode=self.mode,
            anti_aliasing=self.anti_aliasing,
            interpolation=self.interpolation
        )


#import collections
import collections.abc as collections
import cv2
import math
import numpy as np
import numbers
import random
import torch
from imgaug import augmenters as iaa
import matplotlib
import matplotlib.cm
import mono.utils.weather_aug_utils as wa

"""
Provides a set of Pytorch transforms that use OpenCV instead of PIL (Pytorch default)
for image manipulation.
"""

class Compose(object):
    # Composes transforms: transforms.Compose([transforms.RandScale([0.5, 2.0]), transforms.ToTensor()])
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        for t in self.transforms:
            images, labels, intrinsics, cam_models, normals, other_labels, transform_paras = t(images, labels, intrinsics, cam_models, normals, other_labels, transform_paras)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras

class ToTensor(object):
    # Converts numpy.ndarray (H x W x C) to a torch.FloatTensor of shape (C x H x W).
    def __init__(self,  **kwargs):
        return
    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        if not isinstance(images, list) or not isinstance(labels, list) or not isinstance(intrinsics, list):
            raise (RuntimeError("transform.ToTensor() only handle inputs/labels/intrinsics lists."))
        if len(images) != len(intrinsics):
            raise (RuntimeError("Numbers of images and intrinsics are not matched."))
        if not isinstance(images[0], np.ndarray) or not isinstance(labels[0], np.ndarray):
            raise (RuntimeError("transform.ToTensor() only handle np.ndarray for the input and label."
                                "[eg: data readed by cv2.imread()].\n"))
        if  not isinstance(intrinsics[0], list):
            raise (RuntimeError("transform.ToTensor() only handle list for the camera intrinsics"))

        if len(images[0].shape) > 3 or len(images[0].shape) < 2:
            raise (RuntimeError("transform.ToTensor() only handle image(np.ndarray) with 3 dims or 2 dims.\n"))
        if len(labels[0].shape) > 3 or len(labels[0].shape) < 2:
            raise (RuntimeError("transform.ToTensor() only handle label(np.ndarray) with 3 dims or 2 dims.\n"))

        if len(intrinsics[0]) >4 or len(intrinsics[0]) < 3:
            raise (RuntimeError("transform.ToTensor() only handle intrinsic(list) with 3 sizes or 4 sizes.\n"))
        
        for i, img in enumerate(images):
            if len(img.shape) == 2:
                img = np.expand_dims(img, axis=2)
            images[i] = torch.from_numpy(img.transpose((2, 0, 1))).float()
        for i, lab in enumerate(labels):
            if len(lab.shape) == 2:
                lab = np.expand_dims(lab, axis=0)
            labels[i] = torch.from_numpy(lab).float()
        for i, intrinsic in enumerate(intrinsics):
            if len(intrinsic) == 3:
                intrinsic = [intrinsic[0],] + intrinsic
            intrinsics[i] = torch.tensor(intrinsic, dtype=torch.float)
        if cam_models is not None:
            for i, cam_model in enumerate(cam_models):
                cam_models[i] = torch.from_numpy(cam_model.transpose((2, 0, 1))).float() if cam_model is not None else None
        if normals is not None:
            for i, normal in enumerate(normals):
                normals[i] = torch.from_numpy(normal.transpose((2, 0, 1))).float()
        if other_labels is not None:
            for i, lab in enumerate(other_labels):
                if len(lab.shape) == 2:
                    lab = np.expand_dims(lab, axis=0)
                other_labels[i] = torch.from_numpy(lab).float()
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras

class Normalize(object):
    # Normalize tensor with mean and standard deviation along channel: channel = (channel - mean) / std
    def __init__(self, mean, std=None, **kwargs):
        if std is None:
            assert len(mean) > 0
        else:
            assert len(mean) == len(std)
        self.mean = torch.tensor(mean).float()[:, None, None]
        self.std = torch.tensor(std).float()[:, None, None] if std is not None \
            else torch.tensor([1.0, 1.0, 1.0]).float()[:, None, None]

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        # if self.std is None:
        #     # for t, m in zip(image, self.mean):
        #     #     t.sub(m)
        #     image = image - self.mean
        #     if ref_images is not None:
        #         for i, ref_i in enumerate(ref_images):
        #             ref_images[i] =  ref_i - self.mean
        # else:
        #     # for t, m, s in zip(image, self.mean, self.std):
        #     #     t.sub(m).div(s)
        #     image = (image - self.mean) / self.std
        #     if ref_images is not None:
        #         for i, ref_i in enumerate(ref_images):
        #             ref_images[i] =  (ref_i - self.mean) / self.std
        for i, img in enumerate(images):
            img = torch.div((img - self.mean), self.std)
            images[i] = img
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras

class ResizeCanonical(object):
    """
    Resize the input to the canonical space first, then resize the input with random sampled size.
    In the first stage, we assume the distance holds while the camera model varies. 
    In the second stage, we aim to simulate the observation in different distance. The camera will move along the optical axis.
    Args:
        images: list of RGB images.
        labels: list of depth/disparity labels.
        other labels: other labels, such as instance segmentations, semantic segmentations...
    """
    def __init__(self, **kwargs):
        self.ratio_range = kwargs['ratio_range']
        self.canonical_focal = kwargs['focal_length']
        self.crop_size = kwargs['crop_size']
    
    def random_on_canonical_transform(self, image, label, intrinsic, cam_model, to_random_ratio):
        ori_h, ori_w, _ = image.shape
        ori_focal = (intrinsic[0] + intrinsic[1]) / 2.0

        to_canonical_ratio = self.canonical_focal / ori_focal
        to_scale_ratio = to_random_ratio
        resize_ratio = to_canonical_ratio * to_random_ratio
        reshape_h = int(ori_h * resize_ratio + 0.5)
        reshape_w = int(ori_w * resize_ratio + 0.5)

        image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
        if intrinsic is not None:
            intrinsic = [self.canonical_focal, self.canonical_focal, intrinsic[2]*resize_ratio, intrinsic[3]*resize_ratio]
        if label is not None:
            # number of other labels may be less than that of image
            label = cv2.resize(label, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
            # scale the label and camera intrinsics
            label = label / to_scale_ratio
        
        if cam_model is not None:
            # Should not directly resize the cam_model.
            # Camera model should be resized in 'to canonical' stage, while it holds in 'random resizing' stage. 
            # cam_model = cv2.resize(cam_model, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
            cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
        
        return image, label, intrinsic, cam_model, to_scale_ratio       
    
    def random_on_crop_transform(self, image, label, intrinsic, cam_model, to_random_ratio):
        ori_h, ori_w, _ = image.shape
        crop_h, crop_w = self.crop_size
        ori_focal = (intrinsic[0] + intrinsic[1]) / 2.0

        to_canonical_ratio = self.canonical_focal / ori_focal
        
        # random resize based on the last crop size
        proposal_reshape_h = int(crop_h * to_random_ratio + 0.5) 
        proposal_reshape_w = int(crop_w * to_random_ratio + 0.5)
        resize_ratio_h = proposal_reshape_h / ori_h
        resize_ratio_w = proposal_reshape_w / ori_w
        resize_ratio = min(resize_ratio_h, resize_ratio_w) # resize based on the long edge
        reshape_h = int(ori_h * resize_ratio + 0.5)
        reshape_w = int(ori_w * resize_ratio + 0.5)
        
        to_scale_ratio = resize_ratio / to_canonical_ratio        

        image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
        if intrinsic is not None:
            intrinsic = [self.canonical_focal, self.canonical_focal, intrinsic[2]*resize_ratio, intrinsic[3]*resize_ratio]
        if label is not None:
            # number of other labels may be less than that of image
            label = cv2.resize(label, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
            # scale the label and camera intrinsics
            label = label / to_scale_ratio
        
        if cam_model is not None:
            # Should not directly resize the cam_model.
            # Camera model should be resized in 'to canonical' stage, while it holds in 'random resizing' stage. 
            # cam_model = cv2.resize(cam_model, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
            cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
        return image, label, intrinsic, cam_model, to_scale_ratio       

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        assert len(images[0].shape) == 3 and len(labels[0].shape) == 2
        assert labels[0].dtype == np.float
        target_focal = (intrinsics[0][0] + intrinsics[0][1]) / 2.0
        target_to_canonical_ratio = self.canonical_focal / target_focal
        target_img_shape = images[0].shape
        to_random_ratio = random.uniform(self.ratio_range[0], self.ratio_range[1])
        to_scale_ratio = 0.0
        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i] if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model, to_scale_ratio = self.random_on_canonical_transform(
                img, label, intrinsic, cam_model, to_random_ratio)
                
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic
            if cam_model is not None:
                cam_models[i] = cam_model 

        if normals != None:
            reshape_h, reshape_w, _ = images[0].shape
            for i, normal in enumerate(normals):
                normals[i] = cv2.resize(normal, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
                
        if other_labels != None: 
            # other labels are like semantic segmentations, instance segmentations, instance planes segmentations...           
            #resize_ratio = target_to_canonical_ratio * to_scale_ratio
            #reshape_h = int(target_img_shape[0] * resize_ratio + 0.5)
            #reshape_w = int(target_img_shape[1] * resize_ratio + 0.5)
            reshape_h, reshape_w, _ = images[0].shape
            for i, other_label_i in enumerate(other_labels):
                other_labels[i] = cv2.resize(other_label_i, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
        
        if transform_paras is not None:
            transform_paras.update(label_scale_factor = 1.0/to_scale_ratio)

        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras


class LabelScaleCononical(object):
    """
    To solve the ambiguity observation for the mono branch, i.e. different focal length (object size) with the same depth, cameras are
    mapped to a cononical space. To mimic this, we set the focal length to a canonical one and scale the depth value. NOTE: resize the image based on the ratio can also solve this ambiguity.
    Args:
        images: list of RGB images.
        labels: list of depth/disparity labels.
        other labels: other labels, such as instance segmentations, semantic segmentations...
    """
    def __init__(self, **kwargs):
        self.canonical_focal = kwargs['focal_length']
    
    def _get_scale_ratio(self, intrinsic):
        target_focal_x = intrinsic[0]
        label_scale_ratio = self.canonical_focal / target_focal_x
        pose_scale_ratio = 1.0
        return label_scale_ratio, pose_scale_ratio
        
    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        assert len(images[0].shape) == 3 and len(labels[0].shape) == 2
        #assert labels[0].dtype == np.float

        label_scale_ratio = None
        pose_scale_ratio = None
        
        for i in range(len(intrinsics)): 
            img_i = images[i]
            label_i = labels[i] if  i < len(labels) else None
            intrinsic_i = intrinsics[i].copy()
            cam_model_i = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            
            label_scale_ratio, pose_scale_ratio = self._get_scale_ratio(intrinsic_i)
            
            # adjust the focal length, map the current camera to the canonical space
            intrinsics[i] = [intrinsic_i[0]*label_scale_ratio, intrinsic_i[1]*label_scale_ratio, intrinsic_i[2], intrinsic_i[3]]
            
            # scale the label to the canonical space
            if label_i is not None:
                labels[i] = label_i * label_scale_ratio
                        
            if cam_model_i is not None:
                # As the focal length is adjusted (canonical focal length), the camera model should be re-built.
                ori_h, ori_w, _ = img_i.shape
                cam_models[i] = build_camera_model(ori_h, ori_w, intrinsics[i])
            
        
        if transform_paras is not None:
            transform_paras.update(label_scale_factor = label_scale_ratio)

        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras



class ResizeKeepRatio(object):
    """
    Resize and pad to a given size. Hold the aspect ratio.
    This resizing assumes that the camera model remains unchanged.
    Args:
        resize_size: predefined output size.
    """
    def __init__(self, resize_size, padding=None, ignore_label=-1, **kwargs):
        if isinstance(resize_size, int):
            self.resize_h = resize_size
            self.resize_w = resize_size
        elif isinstance(resize_size, collections.Iterable) and len(resize_size) == 2 \
                and isinstance(resize_size[0], int) and isinstance(resize_size[1], int) \
                and resize_size[0] > 0 and resize_size[1] > 0:
            self.resize_h = resize_size[0]
            self.resize_w = resize_size[1]
        else:
            raise (RuntimeError("crop size error.\n"))
        if padding is None:
            self.padding = padding
        elif isinstance(padding, list):
            if all(isinstance(i, numbers.Number) for i in padding):
                self.padding = padding
            else:
                raise (RuntimeError("padding in Crop() should be a number list\n"))
            if len(padding) != 3:
                raise (RuntimeError("padding channel is not equal with 3\n"))
        else:
            raise (RuntimeError("padding in Crop() should be a number list\n"))
        if isinstance(ignore_label, int):
            self.ignore_label = ignore_label
        else:
            raise (RuntimeError("ignore_label should be an integer number\n"))
        self.crop_size = kwargs['crop_size']
        self.canonical_focal = kwargs['focal_length']

    def main_data_transform(self, image, label, intrinsic, cam_model, resize_ratio, padding, to_scale_ratio):
        """
        Resize data first and then do the padding.
        'label' will be scaled.
        """
        h, w, _ = image.shape
        reshape_h = int(resize_ratio * h)
        reshape_w = int(resize_ratio * w)

        pad_h, pad_w, pad_h_half, pad_w_half = padding
        
        # resize
        image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
        # padding
        image = cv2.copyMakeBorder(
            image, 
            pad_h_half, 
            pad_h - pad_h_half, 
            pad_w_half, 
            pad_w - pad_w_half, 
            cv2.BORDER_CONSTANT, 
            value=self.padding)

        if label is not None:
            # label = cv2.resize(label, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
            label = resize_depth_preserve(label, (reshape_h, reshape_w))
            label = cv2.copyMakeBorder(
                label, 
                pad_h_half, 
                pad_h - pad_h_half, 
                pad_w_half, 
                pad_w - pad_w_half, 
                cv2.BORDER_CONSTANT, 
                value=self.ignore_label)
            # scale the label
            label = label / to_scale_ratio
        
        # Resize, adjust principle point
        if intrinsic is not None:
            intrinsic[2] = intrinsic[2] * resize_ratio
            intrinsic[3] = intrinsic[3] * resize_ratio

        if cam_model is not None:
            #cam_model = cv2.resize(cam_model, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
            cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
            cam_model = cv2.copyMakeBorder(
                cam_model, 
                pad_h_half, 
                pad_h - pad_h_half, 
                pad_w_half, 
                pad_w - pad_w_half, 
                cv2.BORDER_CONSTANT, 
                value=self.ignore_label)

        # Pad, adjust the principle point
        if intrinsic is not None:
            intrinsic[2] = intrinsic[2] + pad_w_half
            intrinsic[3] = intrinsic[3] + pad_h_half
        return image, label, intrinsic, cam_model
    
    def get_label_scale_factor(self, image, intrinsic, resize_ratio):
        ori_h, ori_w, _ = image.shape
        crop_h, crop_w = self.crop_size
        ori_focal = (intrinsic[0] + intrinsic[1]) / 2.0 #intrinsic[0] #

        to_canonical_ratio = self.canonical_focal / ori_focal
        to_scale_ratio = resize_ratio / to_canonical_ratio 
        return to_scale_ratio
        
    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        target_h, target_w, _ = images[0].shape
        resize_ratio_h = self.resize_h / target_h
        resize_ratio_w = self.resize_w / target_w
        resize_ratio = min(resize_ratio_h, resize_ratio_w)
        reshape_h = int(resize_ratio * target_h)
        reshape_w = int(resize_ratio * target_w)
        pad_h = max(self.resize_h - reshape_h, 0)
        pad_w = max(self.resize_w - reshape_w, 0)
        pad_h_half = int(pad_h / 2)
        pad_w_half = int(pad_w / 2)
        
        pad_info = [pad_h, pad_w, pad_h_half, pad_w_half]
        to_scale_ratio = self.get_label_scale_factor(images[0], intrinsics[0], resize_ratio)

        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i] if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model = self.main_data_transform(
                img, label, intrinsic, cam_model, resize_ratio, pad_info, to_scale_ratio)
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic
            if cam_model is not None:
                cam_models[i] = cam_model 
        
        if normals is not None:
            for i, normal in enumerate(normals):
                normal =  cv2.resize(normal, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
                # pad
                normals[i] =  cv2.copyMakeBorder(
                    normal, 
                    pad_h_half, 
                    pad_h - pad_h_half, 
                    pad_w_half, 
                    pad_w - pad_w_half, 
                    cv2.BORDER_CONSTANT, 
                    value=0)

        if other_labels is not None:
            
            for i, other_lab in enumerate(other_labels):
                # resize
                other_lab =  cv2.resize(other_lab, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
                # pad
                other_labels[i] =  cv2.copyMakeBorder(
                    other_lab, 
                    pad_h_half, 
                    pad_h - pad_h_half, 
                    pad_w_half, 
                    pad_w - pad_w_half, 
                    cv2.BORDER_CONSTANT, 
                    value=self.ignore_label)


        if transform_paras is not None:
            transform_paras.update(pad=[pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half])
            if 'label_scale_factor' in transform_paras:
                transform_paras['label_scale_factor'] = transform_paras['label_scale_factor'] * 1.0 / to_scale_ratio
            else:
                transform_paras.update(label_scale_factor=1.0/to_scale_ratio)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras
    
class KeepResizeCanoSize(object):
    """
    Resize and pad to a given size. Hold the aspect ratio.
    This resizing assumes that the camera model remains unchanged.
    Args:
        resize_size: predefined output size.
    """
    def __init__(self, resize_size, padding=None, ignore_label=-1, **kwargs):
        if isinstance(resize_size, int):
            self.resize_h = resize_size
            self.resize_w = resize_size
        elif isinstance(resize_size, collections.Iterable) and len(resize_size) == 2 \
                and isinstance(resize_size[0], int) and isinstance(resize_size[1], int) \
                and resize_size[0] > 0 and resize_size[1] > 0:
            self.resize_h = resize_size[0]
            self.resize_w = resize_size[1]
        else:
            raise (RuntimeError("crop size error.\n"))
        if padding is None:
            self.padding = padding
        elif isinstance(padding, list):
            if all(isinstance(i, numbers.Number) for i in padding):
                self.padding = padding
            else:
                raise (RuntimeError("padding in Crop() should be a number list\n"))
            if len(padding) != 3:
                raise (RuntimeError("padding channel is not equal with 3\n"))
        else:
            raise (RuntimeError("padding in Crop() should be a number list\n"))
        if isinstance(ignore_label, int):
            self.ignore_label = ignore_label
        else:
            raise (RuntimeError("ignore_label should be an integer number\n"))
        self.crop_size = kwargs['crop_size']
        self.canonical_focal = kwargs['focal_length']

    def main_data_transform(self, image, label, intrinsic, cam_model, resize_ratio, padding, to_scale_ratio):
        """
        Resize data first and then do the padding.
        'label' will be scaled.
        """
        h, w, _ = image.shape
        reshape_h = int(resize_ratio * h)
        reshape_w = int(resize_ratio * w)

        pad_h, pad_w, pad_h_half, pad_w_half = padding
        
        # resize
        image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
        # padding
        image = cv2.copyMakeBorder(
            image, 
            pad_h_half, 
            pad_h - pad_h_half, 
            pad_w_half, 
            pad_w - pad_w_half, 
            cv2.BORDER_CONSTANT, 
            value=self.padding)

        if label is not None:
            # label = cv2.resize(label, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
            label = resize_depth_preserve(label, (reshape_h, reshape_w))
            label = cv2.copyMakeBorder(
                label, 
                pad_h_half, 
                pad_h - pad_h_half, 
                pad_w_half, 
                pad_w - pad_w_half, 
                cv2.BORDER_CONSTANT, 
                value=self.ignore_label)
            # scale the label
            label = label / to_scale_ratio
        
        # Resize, adjust principle point
        if intrinsic is not None:
            intrinsic[2] = intrinsic[2] * resize_ratio
            intrinsic[3] = intrinsic[3] * resize_ratio

        if cam_model is not None:
            #cam_model = cv2.resize(cam_model, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
            cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
            cam_model = cv2.copyMakeBorder(
                cam_model, 
                pad_h_half, 
                pad_h - pad_h_half, 
                pad_w_half, 
                pad_w - pad_w_half, 
                cv2.BORDER_CONSTANT, 
                value=self.ignore_label)

        # Pad, adjust the principle point
        if intrinsic is not None:
            intrinsic[2] = intrinsic[2] + pad_w_half
            intrinsic[3] = intrinsic[3] + pad_h_half
        return image, label, intrinsic, cam_model
    
    # def get_label_scale_factor(self, image, intrinsic, resize_ratio):
    #     ori_h, ori_w, _ = image.shape
    #     crop_h, crop_w = self.crop_size
    #     ori_focal = intrinsic[0] #(intrinsic[0] + intrinsic[1]) / 2.0

    #     to_canonical_ratio = self.canonical_focal / ori_focal
    #     to_scale_ratio = resize_ratio / to_canonical_ratio 
    #     return to_scale_ratio
        
    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        target_h, target_w, _ = images[0].shape
        ori_focal = intrinsics[0][0]
        to_canonical_ratio = self.canonical_focal / ori_focal 
        
        resize_ratio = to_canonical_ratio
        reshape_h = int(resize_ratio * target_h)
        reshape_w = int(resize_ratio * target_w)
        
        pad_h = 32 - reshape_h % 32
        pad_w = 32 - reshape_w % 32
        pad_h_half = int(pad_h / 2)
        pad_w_half = int(pad_w / 2)
        
        pad_info = [pad_h, pad_w, pad_h_half, pad_w_half]
        to_scale_ratio = 1.0

        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i] if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model = self.main_data_transform(
                img, label, intrinsic, cam_model, resize_ratio, pad_info, to_scale_ratio)
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic
            if cam_model is not None:
                cam_models[i] = cam_model 
        
        if normals is not None:
            
            for i, normal in enumerate(normals):
                # resize
                normal =  cv2.resize(normal, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
                # pad
                normals[i] =  cv2.copyMakeBorder(
                    normal, 
                    pad_h_half, 
                    pad_h - pad_h_half, 
                    pad_w_half, 
                    pad_w - pad_w_half, 
                    cv2.BORDER_CONSTANT, 
                    value=0)

        if other_labels is not None:
            
            for i, other_lab in enumerate(other_labels):
                # resize
                other_lab =  cv2.resize(other_lab, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
                # pad
                other_labels[i] =  cv2.copyMakeBorder(
                    other_lab, 
                    pad_h_half, 
                    pad_h - pad_h_half, 
                    pad_w_half, 
                    pad_w - pad_w_half, 
                    cv2.BORDER_CONSTANT, 
                    value=self.ignore_label)


        if transform_paras is not None:
            transform_paras.update(pad=[pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half])
            if 'label_scale_factor' in transform_paras:
                transform_paras['label_scale_factor'] = transform_paras['label_scale_factor'] * 1.0 / to_scale_ratio
            else:
                transform_paras.update(label_scale_factor=1.0/to_scale_ratio)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras


class RandomCrop(object):
    """Crops the given ndarray image (H*W*C or H*W).
    Args:
        size (sequence or int): Desired output size of the crop. If size is an
        int instead of sequence like (h, w), a square crop (size, size) is made.
    """
    def __init__(self, crop_size, crop_type='center', padding=None, ignore_label=-1, **kwargs):
        if isinstance(crop_size, int):
            self.crop_h = crop_size
            self.crop_w = crop_size
        elif isinstance(crop_size, collections.Iterable) and len(crop_size) == 2 \
                and isinstance(crop_size[0], int) and isinstance(crop_size[1], int) \
                and crop_size[0] > 0 and crop_size[1] > 0:
            self.crop_h = crop_size[0]
            self.crop_w = crop_size[1]
        else:
            raise (RuntimeError("crop size error.\n"))
        if crop_type == 'center' or crop_type == 'rand' or crop_type=='rand_in_field':
            self.crop_type = crop_type
        else:
            raise (RuntimeError("crop type error: rand | center | rand_in_field \n"))
        if padding is None:
            self.padding = padding
        elif isinstance(padding, list):
            if all(isinstance(i, numbers.Number) for i in padding):
                self.padding = padding
            else:
                raise (RuntimeError("padding in Crop() should be a number list\n"))
            if len(padding) != 3:
                raise (RuntimeError("padding channel is not equal with 3\n"))
        else:
            raise (RuntimeError("padding in Crop() should be a number list\n"))
        if isinstance(ignore_label, int):
            self.ignore_label = ignore_label
        else:
            raise (RuntimeError("ignore_label should be an integer number\n"))
        

    def cal_padding_paras(self, h, w):
        # padding if current size is not satisfied
        pad_h = max(self.crop_h - h, 0)
        pad_w = max(self.crop_w - w, 0)
        pad_h_half = int(pad_h / 2)
        pad_w_half = int(pad_w / 2)
        return pad_h, pad_w, pad_h_half, pad_w_half

    def cal_cropping_paras(self, h, w, intrinsic):
        u0 = intrinsic[2]
        v0 = intrinsic[3]
        if self.crop_type == 'rand':
            h_min = 0
            h_max = h - self.crop_h
            w_min = 0
            w_max = w - self.crop_w 
        elif self.crop_type == 'center':
            h_min = (h - self.crop_h) / 2
            h_max = (h - self.crop_h) / 2
            w_min = (w - self.crop_w) / 2
            w_max = (w - self.crop_w) / 2
        else: # rand in field
            h_min = min(max(0, v0 - 0.75*self.crop_h), h-self.crop_h)
            h_max = min(max(v0 - 0.25*self.crop_h, 0), h-self.crop_h)
            w_min = min(max(0, u0 - 0.75*self.crop_w), w-self.crop_w)
            w_max = min(max(u0 - 0.25*self.crop_w, 0), w-self.crop_w)
        
        h_off = random.randint(int(h_min), int(h_max))
        w_off = random.randint(int(w_min), int(w_max))
        return h_off, w_off
    
    def main_data_transform(self, image, label, intrinsic, cam_model,
        pad_h, pad_w, pad_h_half, pad_w_half, h_off, w_off):

        # padding if current size is not satisfied
        if pad_h > 0 or pad_w > 0:
            if self.padding is None:
                raise (RuntimeError("depthtransform.Crop() need padding while padding argument is None\n"))
            image = cv2.copyMakeBorder(image, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=self.padding)
            if label is not None:
                label = cv2.copyMakeBorder(label, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=self.ignore_label)
            if cam_model is not None:
                cam_model = cv2.copyMakeBorder(cam_model, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=self.ignore_label)
        
        # cropping
        image = image[h_off:h_off+self.crop_h, w_off:w_off+self.crop_w]
        if label is not None:
            label = label[h_off:h_off+self.crop_h, w_off:w_off+self.crop_w]
        if cam_model is not None:
            cam_model = cam_model[h_off:h_off+self.crop_h, w_off:w_off+self.crop_w]

        if intrinsic is not None:
            intrinsic[2] = intrinsic[2] + pad_w_half - w_off
            intrinsic[3] = intrinsic[3] + pad_h_half - h_off    
        return image, label, intrinsic, cam_model

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        if 'random_crop_size' in transform_paras and transform_paras['random_crop_size'] is not None \
            and (transform_paras['random_crop_size'][0] + transform_paras['random_crop_size'][1] > 500):
            self.crop_h = int(transform_paras['random_crop_size'][0].item())
            self.crop_w = int(transform_paras['random_crop_size'][1].item())
        target_img = images[0]
        target_h, target_w, _ = target_img.shape
        target_intrinsic = intrinsics[0]
        pad_h, pad_w, pad_h_half, pad_w_half = self.cal_padding_paras(target_h, target_w)
        h_off, w_off = self.cal_cropping_paras(target_h+pad_h, target_w+pad_w, target_intrinsic)

        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i].copy() if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model = self.main_data_transform(
                img, label, intrinsic, cam_model,
                pad_h, pad_w, pad_h_half, pad_w_half, h_off, w_off)
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic
            if cam_model is not None:
                cam_models[i] = cam_model 
        pad=[pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half]
        if normals is not None:           
            for i, normal  in enumerate(normals):
                # padding if current size is not satisfied
                normal = cv2.copyMakeBorder(normal, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=0)
                normals[i] = normal[h_off:h_off+self.crop_h, w_off:w_off+self.crop_w]
        if other_labels is not None:           
            for i, other_lab  in enumerate(other_labels):
                # padding if current size is not satisfied
                other_lab = cv2.copyMakeBorder(other_lab, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=self.ignore_label)
                other_labels[i] = other_lab[h_off:h_off+self.crop_h, w_off:w_off+self.crop_w]
        if transform_paras is not None:
            transform_paras.update(dict(pad=pad))
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras
    

class RandomResize(object):
    """
    Random resize the image. During this process, the camera model is hold, and thus the depth label is scaled.
    Args:
        images: list of RGB images.
        labels: list of depth/disparity labels.
        other labels: other labels, such as instance segmentations, semantic segmentations...
    """
    def __init__(self, ratio_range=(0.85, 1.15), prob=0.5, is_lidar=True, **kwargs):
        self.ratio_range = ratio_range
        self.is_lidar = is_lidar
        self.prob = prob
    
    def random_resize(self, image, label, intrinsic, cam_model, to_random_ratio):
        ori_h, ori_w, _ = image.shape
        
        resize_ratio = to_random_ratio
        label_scale_ratio = 1.0 / resize_ratio
        reshape_h = int(ori_h * resize_ratio + 0.5)
        reshape_w = int(ori_w * resize_ratio + 0.5)

        image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
        if intrinsic is not None:
            intrinsic = [intrinsic[0], intrinsic[1], intrinsic[2]*resize_ratio, intrinsic[3]*resize_ratio]
        if label is not None:
            if self.is_lidar:
                label = resize_depth_preserve(label, (reshape_h, reshape_w))
            else:
                label = cv2.resize(label, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
            # scale the label
            label = label * label_scale_ratio
        
        if cam_model is not None:
            # Should not directly resize the cam_model.
            # Camera model should be resized in 'to canonical' stage, while it holds in 'random resizing' stage. 
            # cam_model = cv2.resize(cam_model, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
            cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
        
        return image, label, intrinsic, cam_model, label_scale_ratio       

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        assert len(images[0].shape) == 3 and len(labels[0].shape) == 2
        assert labels[0].dtype == np.float
        # target_focal = (intrinsics[0][0] + intrinsics[0][1]) / 2.0
        # target_to_canonical_ratio = self.canonical_focal / target_focal
        # target_img_shape = images[0].shape
        prob = random.uniform(0, 1)
        if prob < self.prob:
            to_random_ratio = random.uniform(self.ratio_range[0], self.ratio_range[1])
        else:
            to_random_ratio = 1.0
        label_scale_ratio = 0.0
        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i].copy() if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model, label_scale_ratio = self.random_resize(
                img, label, intrinsic, cam_model, to_random_ratio)
                
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic.copy()
            if cam_model is not None:
                cam_models[i] = cam_model 

        if normals != None: 
            reshape_h, reshape_w, _ = images[0].shape
            for i, norm in enumerate(normals):
                normals[i] = cv2.resize(norm, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)


        if other_labels != None: 
            # other labels are like semantic segmentations, instance segmentations, instance planes segmentations...           
            #resize_ratio = target_to_canonical_ratio * to_scale_ratio
            #reshape_h = int(target_img_shape[0] * resize_ratio + 0.5)
            #reshape_w = int(target_img_shape[1] * resize_ratio + 0.5)
            reshape_h, reshape_w, _ = images[0].shape
            for i, other_label_i in enumerate(other_labels):
                other_labels[i] = cv2.resize(other_label_i, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
        
        if transform_paras is not None:
            if 'label_scale_factor' in transform_paras:
                transform_paras['label_scale_factor'] = transform_paras['label_scale_factor'] * label_scale_ratio
            else:
                transform_paras.update(label_scale_factor = label_scale_ratio)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras

class RandomEdgeMask(object):
    """
    Random mask the input and labels.
    Args:
        images: list of RGB images.
        labels: list of depth/disparity labels.
        other labels: other labels, such as instance segmentations, semantic segmentations...
    """
    def __init__(self, mask_maxsize=32, prob=0.5, rgb_invalid=[0,0,0], label_invalid=-1,**kwargs):
        self.mask_maxsize = mask_maxsize
        self.prob = prob
        self.rgb_invalid = rgb_invalid
        self.label_invalid = label_invalid
    
    def mask_edge(self, image, mask_edgesize, mask_value):
        H, W = image.shape[0], image.shape[1]
        # up
        image[0:mask_edgesize[0], :, ...] = mask_value 
        # down
        image[H-mask_edgesize[1]:H, :, ...] = mask_value
        # left
        image[:, 0:mask_edgesize[2], ...] = mask_value
        # right
        image[:, W-mask_edgesize[3]:W, ...] = mask_value
        
        return image

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        assert len(images[0].shape) == 3 and len(labels[0].shape) == 2
        assert labels[0].dtype == np.float
        
        prob = random.uniform(0, 1)
        if prob > self.prob:
            return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras
        
        mask_edgesize = random.sample(range(self.mask_maxsize), 4) #[up, down, left, right]
        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            img = self.mask_edge(img, mask_edgesize, self.rgb_invalid)
                
            images[i] = img
            if label is not None:
                label = self.mask_edge(label, mask_edgesize, self.label_invalid)
                labels[i] = label
        
        if normals != None:       
            for i, normal in enumerate(normals):
                normals[i] = self.mask_edge(normal, mask_edgesize, mask_value=0)

        if other_labels != None: 
            # other labels are like semantic segmentations, instance segmentations, instance planes segmentations...           
            for i, other_label_i in enumerate(other_labels):
                other_labels[i] = self.mask_edge(other_label_i, mask_edgesize, self.label_invalid)

        if transform_paras is not None:
            pad = transform_paras['pad'] if 'pad' in transform_paras else [0,0,0,0]
            new_pad = [max(mask_edgesize[0], pad[0]), max(mask_edgesize[1], pad[1]), max(mask_edgesize[2], pad[2]), max(mask_edgesize[3], pad[3])]
            transform_paras.update(dict(pad=new_pad)) 
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras


class AdjustSize(object):
    """Crops the given ndarray image (H*W*C or H*W).
    Args:
        size (sequence or int): Desired output size of the crop. If size is an
        int instead of sequence like (h, w), a square crop (size, size) is made.
    """
    def __init__(self, padding=None, ignore_label=-1, **kwargs):
        
        if padding is None:
            self.padding = padding
        elif isinstance(padding, list):
            if all(isinstance(i, numbers.Number) for i in padding):
                self.padding = padding
            else:
                raise (RuntimeError("padding in Crop() should be a number list\n"))
            if len(padding) != 3:
                raise (RuntimeError("padding channel is not equal with 3\n"))
        else:
            raise (RuntimeError("padding in Crop() should be a number list\n"))
        if isinstance(ignore_label, int):
            self.ignore_label = ignore_label
        else:
            raise (RuntimeError("ignore_label should be an integer number\n"))
        
    def get_pad_paras(self, h, w):
        pad_h = 32 - h % 32 if h %32 != 0 else 0
        pad_w = 32 - w % 32 if w %32 != 0 else 0
        pad_h_half = int(pad_h // 2)
        pad_w_half = int(pad_w // 2)
        return pad_h, pad_w, pad_h_half, pad_w_half

    def main_data_transform(self, image, label, intrinsic, cam_model):
        h, w, _ = image.shape
        pad_h, pad_w, pad_h_half, pad_w_half = self.get_pad_paras(h=h, w=w)
        if pad_h > 0 or pad_w > 0:
            if self.padding is None:
                raise (RuntimeError("depthtransform.Crop() need padding while padding argument is None\n"))
            image = cv2.copyMakeBorder(image, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=self.padding)
            if label is not None:
                label = cv2.copyMakeBorder(label, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=self.ignore_label)
            if cam_model is not None:
                cam_model = cv2.copyMakeBorder(cam_model, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=self.ignore_label)

        if intrinsic is not None:
            intrinsic[2] = intrinsic[2] + pad_w_half
            intrinsic[3] = intrinsic[3] + pad_h_half
        pad=[pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half]
        return image, label, intrinsic, cam_model, pad
    
    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        target_img = images[0]
        target_h, target_w, _ = target_img.shape
        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i] if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model, pad = self.main_data_transform(
                img, label, intrinsic, cam_model)
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic
            if cam_model is not None:
                cam_models[i] = cam_model 

            if transform_paras is not None:
                transform_paras.update(dict(pad=pad))
        if normals is not None:
            pad_h, pad_w, pad_h_half, pad_w_half = self.get_pad_paras(h=target_h, w=target_w)
            for i, normal  in enumerate(normals):
                normals[i] = cv2.copyMakeBorder(normal, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=0)

        if other_labels is not None:
            pad_h, pad_w, pad_h_half, pad_w_half = self.get_pad_paras(h=target_h, w=target_w)
            for i, other_lab  in enumerate(other_labels):
                other_labels[i] = cv2.copyMakeBorder(other_lab, pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half, cv2.BORDER_CONSTANT, value=self.ignore_label)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras


class RandomHorizontalFlip(object):
    def __init__(self, prob=0.5, **kwargs):
        self.p = prob

    def main_data_transform(self, image, label, intrinsic, cam_model, rotate):
        if rotate:
            image = cv2.flip(image, 1)
            if label is not None:
                label = cv2.flip(label, 1)
            if intrinsic is not None:
                h, w, _ = image.shape
                intrinsic[2] = w - intrinsic[2]
                intrinsic[3] = h - intrinsic[3]
            if cam_model is not None:
                cam_model = cv2.flip(cam_model, 1)
                cam_model[:, :, 0] = cam_model[:, :, 0] * -1
                cam_model[:, :, 2] = cam_model[:, :, 2] * -1
        return image, label, intrinsic, cam_model
    
    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        rotate = random.random() > self.p 

        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i] if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model = self.main_data_transform(
                img, label, intrinsic, cam_model, rotate)
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic
            if cam_model is not None:
                cam_models[i] = cam_model 
        if normals is not None:
            for i, normal in enumerate(normals):
                if rotate:
                    normal = cv2.flip(normal, 1)
                    normal[:, :, 0] = -normal[:, :, 0] # NOTE: check the direction of normal coordinates axis, this is used in https://github.com/baegwangbin/surface_normal_uncertainty
                normals[i] = normal

        if other_labels is not None:
            for i, other_lab in enumerate(other_labels):
                if rotate:
                    other_lab = cv2.flip(other_lab, 1)
                other_labels[i] = other_lab
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras

class RandomBlur(object):
    def __init__(self, 
                 aver_kernal=(2, 10), 
                 motion_kernal=(5, 15), 
                 angle=[-80, 80], 
                 prob=0.3,
                 **kwargs):

        gaussian_blur = iaa.AverageBlur(k=aver_kernal)
        motion_blur = iaa.MotionBlur(k=motion_kernal, angle=angle)
        zoom_blur = iaa.imgcorruptlike.ZoomBlur(severity=1)
        self.prob = prob
        self.blurs = [gaussian_blur, motion_blur, zoom_blur]

    def blur(self, imgs, id):
        blur_mtd = self.blurs[id]
        return blur_mtd(images=imgs)

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        prob = random.random()
        if prob < self.prob:
            id = random.randint(0, len(self.blurs)-1)
            images = self.blur(images, id)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras

class RGBCompresion(object):
    def __init__(self, prob=0.1, compression=(0, 50), **kwargs):
        self.rgb_compress = iaa.Sequential(
            [
                iaa.JpegCompression(compression=compression),
            ],
            random_order=True,
        )
        self.prob = prob

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        if random.random() < self.prob:
            images = self.rgb_compress(images=images)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras


class RGB2BGR(object):
    # Converts image from RGB order to BGR order, for model initialized from Caffe
    def __init__(self,  **kwargs):
        return
    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        for i, img in enumerate(images):
            images[i] = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras


class BGR2RGB(object):
    # Converts image from BGR order to RGB order, for model initialized from Pytorch
    def __init__(self,  **kwargs):
        return
    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        for i, img in enumerate(images):
            images[i] = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras


class PhotoMetricDistortion(object):
    """Apply photometric distortion to image sequentially, every transformation
    is applied with a probability of 0.5. The position of random contrast is in
    second or second to last.
    1. random brightness
    2. random contrast (mode 0)
    3. convert color from BGR to HSV
    4. random saturation
    5. random hue
    6. convert color from HSV to BGR
    7. random contrast (mode 1)
    Args:
        brightness_delta (int): delta of brightness.
        contrast_range (tuple): range of contrast.
        saturation_range (tuple): range of saturation.
        hue_delta (int): delta of hue.
    """

    def __init__(self,
                 brightness_delta=32,
                 contrast_range=(0.5, 1.5),
                 saturation_range=(0.5, 1.5),
                 hue_delta=18,
                 to_gray_prob=0.3,
                 distortion_prob=0.3,
                 **kwargs):
        self.brightness_delta = brightness_delta
        self.contrast_lower, self.contrast_upper = contrast_range
        self.saturation_lower, self.saturation_upper = saturation_range
        self.hue_delta = hue_delta
        self.gray_aug = iaa.Grayscale(alpha=(0.8, 1.0))
        self.to_gray_prob = to_gray_prob
        self.distortion_prob = distortion_prob

    def convert(self, img, alpha=1.0, beta=0.0):
        """Multiple with alpha and add beat with clip."""
        img = img.astype(np.float32) * alpha + beta
        img = np.clip(img, 0, 255)
        return img.astype(np.uint8)

    def brightness(self, img, beta, do):
        """Brightness distortion."""
        if do:
            # beta = random.uniform(-self.brightness_delta,
            #                         self.brightness_delta)
            img = self.convert(
                img,
                beta=beta)
        return img

    def contrast(self, img, alpha, do):
        """Contrast distortion."""
        if do:
            #alpha = random.uniform(self.contrast_lower, self.contrast_upper)
            img = self.convert(
                img,
                alpha=alpha)
        return img

    def saturation(self, img, alpha, do):
        """Saturation distortion."""
        if do:
            # alpha = random.uniform(self.saturation_lower,
            #                         self.saturation_upper)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
            img[:, :, 1] = self.convert(
                img[:, :, 1],
                alpha=alpha)
            img = cv2.cvtColor(img, cv2.COLOR_HSV2BGR)        
        return img

    def hue(self, img, rand_hue, do):
        """Hue distortion."""
        if do:
            # rand_hue = random.randint(-self.hue_delta, self.hue_delta)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
            img[:, :, 0] = (img[:, :, 0].astype(int) + rand_hue) % 180
            img = cv2.cvtColor(img, cv2.COLOR_HSV2BGR)
        return img
    
    def rgb2gray(self, img):
        img = self.gray_aug(image=img)
        return img

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        """Call function to perform photometric distortion on images.
        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Result dict with images distorted.
        """
        brightness_beta = random.uniform(-self.brightness_delta, self.brightness_delta)
        brightness_do = random.random() < self.distortion_prob

        contrast_alpha = random.uniform(self.contrast_lower, self.contrast_upper)
        contrast_do = random.random() < self.distortion_prob

        saturate_alpha = random.uniform(self.saturation_lower, self.saturation_upper)
        saturate_do = random.random() < self.distortion_prob

        rand_hue = random.randint(-self.hue_delta, self.hue_delta)
        rand_hue_do = random.random() < self.distortion_prob

        # mode == 0 --> do random contrast first
        # mode == 1 --> do random contrast last
        mode = 1 if random.random() > 0.5 else 2
        for i, img in enumerate(images):
            if random.random() < self.to_gray_prob:
                img = self.rgb2gray(img)
            else:
                # random brightness
                img = self.brightness(img, brightness_beta, brightness_do)

                if mode == 1:
                    img = self.contrast(img, contrast_alpha, contrast_do)

                # random saturation
                img = self.saturation(img, saturate_alpha, saturate_do)

                # random hue
                img = self.hue(img, rand_hue, rand_hue_do)

                # random contrast
                if mode == 0:
                    img = self.contrast(img, contrast_alpha, contrast_do)
            images[i] = img
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras

class Weather(object):
    """Apply the following weather augmentations to data.
    Args:
        prob (float): probability to enforce the weather augmentation.
    """

    def __init__(self,
                 prob=0.3,
                 **kwargs):
        snow = iaa.FastSnowyLandscape(
            lightness_threshold=[50, 100],
            lightness_multiplier=(1.2, 2)
            )
        cloud = iaa.Clouds()
        fog = iaa.Fog()
        snow_flakes = iaa.Snowflakes(flake_size=(0.2, 0.4), speed=(0.001, 0.03)) #iaa.imgcorruptlike.Snow(severity=2)# 
        rain = iaa.Rain(speed=(0.1, 0.3), drop_size=(0.1, 0.3))
        # rain_drops = RainDrop_Augmentor()
        self.aug_list = [
            snow, cloud, fog, snow_flakes, rain, 
            #wa.add_sun_flare, wa.darken, wa.random_brightness,
        ]
        self.prob = prob
    
    def aug_with_weather(self, imgs, id):
        weather = self.aug_list[id]
        if id <5:
            return weather(images=imgs)
        else:
            return weather(imgs)

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        """Call function to perform photometric distortion on images.
        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Result dict with images distorted.
        """
        if random.random() < self.prob:
            select_id = np.random.randint(0, high=len(self.aug_list))  
            images = self.aug_with_weather(images, select_id)        
        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras
        
    
def resize_depth_preserve(depth, shape):
    """
    Resizes depth map preserving all valid depth pixels
    Multiple downsampled points can be assigned to the same pixel.

    Parameters
    ----------
    depth : np.array [h,w]
        Depth map
    shape : tuple (H,W)
        Output shape

    Returns
    -------
    depth : np.array [H,W,1]
        Resized depth map
    """
    # Store dimensions and reshapes to single column
    depth = np.squeeze(depth)
    h, w = depth.shape
    x = depth.reshape(-1)
    # Create coordinate grid
    uv = np.mgrid[:h, :w].transpose(1, 2, 0).reshape(-1, 2)
    # Filters valid points
    idx = x > 0
    crd, val = uv[idx], x[idx]
    # Downsamples coordinates
    crd[:, 0] = (crd[:, 0] * (shape[0] / h) + 0.5).astype(np.int32)
    crd[:, 1] = (crd[:, 1] * (shape[1] / w) + 0.5).astype(np.int32)
    # Filters points inside image
    idx = (crd[:, 0] < shape[0]) & (crd[:, 1] < shape[1])
    crd, val = crd[idx], val[idx]
    # Creates downsampled depth image and assigns points
    depth = np.zeros(shape)
    depth[crd[:, 0], crd[:, 1]] = val
    # Return resized depth map
    return depth


def gray_to_colormap(img, cmap='rainbow', max_value=None):
    """
    Transfer gray map to matplotlib colormap
    """
    assert img.ndim == 2

    img[img<0] = 0
    mask_invalid = img < 1e-10
    if max_value == None:
        img = img / (img.max() + 1e-8)
    else:
        img = img / (max_value + 1e-8)
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1.1)
    cmap_m = matplotlib.cm.get_cmap(cmap)
    map = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap_m)
    colormap = (map.to_rgba(img)[:, :, :3] * 255).astype(np.uint8)
    colormap[mask_invalid] = 0
    return colormap


class LiDarResizeCanonical(object):
    """
    Resize the input to the canonical space first, then resize the input with random sampled size.
    In the first stage, we assume the distance holds while the camera model varies. 
    In the second stage, we aim to simulate the observation in different distance. The camera will move along the optical axis.
    """
    def __init__(self, **kwargs):
        self.ratio_range = kwargs['ratio_range']
        self.canonical_focal = kwargs['focal_length']
        self.crop_size = kwargs['crop_size']
    
    def random_on_canonical_transform(self, image, label, intrinsic, cam_model, to_random_ratio):
        ori_h, ori_w, _ = image.shape
        ori_focal = (intrinsic[0] + intrinsic[1]) / 2.0

        to_canonical_ratio = self.canonical_focal / ori_focal
        to_scale_ratio = to_random_ratio
        resize_ratio = to_canonical_ratio * to_random_ratio
        reshape_h = int(ori_h * resize_ratio + 0.5)
        reshape_w = int(ori_w * resize_ratio + 0.5)

        image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
        if intrinsic is not None:
            intrinsic = [self.canonical_focal, self.canonical_focal, intrinsic[2]*resize_ratio, intrinsic[3]*resize_ratio]
        if label is not None:
            # number of other labels may be less than that of image
            #label = cv2.resize(label, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
            label = resize_depth_preserve(label, (reshape_h, reshape_w))
            # scale the label and camera intrinsics
            label = label / to_scale_ratio
        
        if cam_model is not None:
            # Should not directly resize the cam_model.
            # Camera model should be resized in 'to canonical' stage, while it holds in 'random resizing' stage. 
            # cam_model = cv2.resize(cam_model, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
            cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
        return image, label, intrinsic, cam_model, to_scale_ratio    
    
    def random_on_crop_transform(self, image, label, intrinsic, cam_model, to_random_ratio):
        ori_h, ori_w, _ = image.shape
        crop_h, crop_w = self.crop_size
        ori_focal = (intrinsic[0] + intrinsic[1]) / 2.0

        to_canonical_ratio = self.canonical_focal / ori_focal
        
        # random resize based on the last crop size
        proposal_reshape_h = int(crop_h * to_random_ratio + 0.5) 
        proposal_reshape_w = int(crop_w * to_random_ratio + 0.5)
        resize_ratio_h = proposal_reshape_h / ori_h
        resize_ratio_w = proposal_reshape_w / ori_w
        resize_ratio = min(resize_ratio_h, resize_ratio_w) # resize based on the long edge
        reshape_h = int(ori_h * resize_ratio + 0.5)
        reshape_w = int(ori_w * resize_ratio + 0.5)
        
        to_scale_ratio = resize_ratio / to_canonical_ratio        

        image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
        if intrinsic is not None:
            intrinsic = [self.canonical_focal, self.canonical_focal, intrinsic[2]*resize_ratio, intrinsic[3]*resize_ratio]
        if label is not None:
            # number of other labels may be less than that of image
            # label = cv2.resize(label, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
            label = resize_depth_preserve(label, (reshape_h, reshape_w))
            # scale the label and camera intrinsics
            label = label / to_scale_ratio
        
        if cam_model is not None:
            # Should not directly resize the cam_model.
            # Camera model should be resized in 'to canonical' stage, while it holds in 'random resizing' stage. 
            # cam_model = cv2.resize(cam_model, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
            cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
        return image, label, intrinsic, cam_model, to_scale_ratio          

    def __call__(self, images, labels, intrinsics, cam_models=None, normals=None, other_labels=None, transform_paras=None):
        assert len(images[0].shape) == 3 and len(labels[0].shape) == 2
        assert labels[0].dtype == np.float
        target_focal = (intrinsics[0][0] + intrinsics[0][1]) / 2.0
        target_to_canonical_ratio = self.canonical_focal / target_focal
        target_img_shape = images[0].shape
        to_random_ratio = random.uniform(self.ratio_range[0], self.ratio_range[1])
        to_scale_ratio = 0
        for i in range(len(images)):
            img = images[i]
            label = labels[i] if i < len(labels) else None
            intrinsic = intrinsics[i] if i < len(intrinsics) else None
            cam_model = cam_models[i] if cam_models is not None and i < len(cam_models) else None
            img, label, intrinsic, cam_model, to_scale_ratio = self.random_on_canonical_transform(
                img, label, intrinsic, cam_model, to_random_ratio)
                
            images[i] = img
            if label is not None:
                labels[i] = label
            if intrinsic is not None:
                intrinsics[i] = intrinsic
            if cam_model is not None:
                cam_models[i] = cam_model 
        if normals != None: 
            reshape_h, reshape_w, _ = images[0].shape
            for i, normal in enumerate(normals):
                normals[i] = cv2.resize(normal, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
                
        if other_labels != None: 
            # other labels are like semantic segmentations, instance segmentations, instance planes segmentations...           
            # resize_ratio = target_to_canonical_ratio * to_random_ratio
            # reshape_h = int(target_img_shape[0] * resize_ratio + 0.5)
            # reshape_w = int(target_img_shape[1] * resize_ratio + 0.5)
            reshape_h, reshape_w, _ = images[0].shape
            for i, other_label_i in enumerate(other_labels):
                other_labels[i] = cv2.resize(other_label_i, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_NEAREST)
              
        if transform_paras is not None:
            transform_paras.update(label_scale_factor = 1.0/to_scale_ratio)

        return images, labels, intrinsics, cam_models, normals, other_labels, transform_paras



def build_camera_model(H : int, W : int, intrinsics : list) -> np.array:
    """
    Encode the camera intrinsic parameters (focal length and principle point) to a 4-channel map. 
    """
    fx, fy, u0, v0 = intrinsics
    f = (fx + fy) / 2.0
    # principle point location
    x_row = np.arange(0, W).astype(np.float32)
    x_row_center_norm = (x_row - u0) / W
    x_center = np.tile(x_row_center_norm, (H, 1)) # [H, W]

    y_col = np.arange(0, H).astype(np.float32) 
    y_col_center_norm = (y_col - v0) / H
    y_center = np.tile(y_col_center_norm, (W, 1)).T

    # FoV
    fov_x = np.arctan(x_center / (f / W))
    fov_y =  np.arctan(y_center/ (f / H))

    cam_model = np.stack([x_center, y_center, fov_x, fov_y], axis=2)
    return cam_model


if __name__ == '__main__':
    img = cv2.imread('/mnt/mldb/raw/62b3ed3455e805efcb28c74b/NuScenes/data_test/samples/CAM_FRONT/n008-2018-08-01-15-34-25-0400__CAM_FRONT__1533152214512404.jpg', -1)
    H, W, _ = img.shape
    label = img[:, :, 0]
    intrinsic = [1000, 1000, W//2, H//2]
    for i in range(20):
        weather_aug = Weather(prob=1.0)
        img_aug,  label, intrinsic, cam_model, ref_images, transform_paras = weather_aug([img, ], [label,], [intrinsic,])
        cv2.imwrite(f'test_aug_{i}.jpg', img_aug[0])
    
    print('Done')



import argparse
import cv2
import glob
import matplotlib
import numpy as np
import os
import torch

from depth_anything_v2.dpt import DepthAnythingV2


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Depth Anything V2')
    
    parser.add_argument('--img-path', type=str)
    parser.add_argument('--input-size', type=int, default=518)
    parser.add_argument('--outdir', type=str, default='./vis_depth')
    
    parser.add_argument('--encoder', type=str, default='vitl', choices=['vits', 'vitb', 'vitl', 'vitg'])
    
    parser.add_argument('--pred-only', dest='pred_only', action='store_true', help='only display the prediction')
    parser.add_argument('--grayscale', dest='grayscale', action='store_true', help='do not apply colorful palette')
    
    args = parser.parse_args()
    
    DEVICE = 'cuda' if torch.cuda.is_available() else 'mps' if torch.backends.mps.is_available() else 'cpu'
    
    model_configs = {
        'vits': {'encoder': 'vits', 'features': 64, 'out_channels': [48, 96, 192, 384]},
        'vitb': {'encoder': 'vitb', 'features': 128, 'out_channels': [96, 192, 384, 768]},
        'vitl': {'encoder': 'vitl', 'features': 256, 'out_channels': [256, 512, 1024, 1024]},
        'vitg': {'encoder': 'vitg', 'features': 384, 'out_channels': [1536, 1536, 1536, 1536]}
    }
    
    depth_anything = DepthAnythingV2(**model_configs[args.encoder])
    depth_anything.load_state_dict(torch.load(f'checkpoints/depth_anything_v2_{args.encoder}.pth', map_location='cpu'))
    depth_anything = depth_anything.to(DEVICE).eval()
    
    if os.path.isfile(args.img_path):
        if args.img_path.endswith('txt'):
            with open(args.img_path, 'r') as f:
                filenames = f.read().splitlines()
        else:
            filenames = [args.img_path]
    else:
        filenames = glob.glob(os.path.join(args.img_path, '**/*'), recursive=True)
    
    os.makedirs(args.outdir, exist_ok=True)
    
    cmap = matplotlib.colormaps.get_cmap('Spectral_r')
    
    for k, filename in enumerate(filenames):
        print(f'Progress {k+1}/{len(filenames)}: {filename}')
        
        raw_image = cv2.imread(filename)
        
        depth = depth_anything.infer_image(raw_image, args.input_size)
        
        depth = (depth - depth.min()) / (depth.max() - depth.min()) * 255.0
        depth = depth.astype(np.uint8)
        
        if args.grayscale:
            depth = np.repeat(depth[..., np.newaxis], 3, axis=-1)
        else:
            depth = (cmap(depth)[:, :, :3] * 255)[:, :, ::-1].astype(np.uint8)
        
        if args.pred_only:
            cv2.imwrite(os.path.join(args.outdir, os.path.splitext(os.path.basename(filename))[0] + '.png'), depth)
        else:
            split_region = np.ones((raw_image.shape[0], 50, 3), dtype=np.uint8) * 255
            combined_result = cv2.hconcat([raw_image, split_region, depth])
            
            cv2.imwrite(os.path.join(args.outdir, os.path.splitext(os.path.basename(filename))[0] + '.png'), combined_result)


from ...utils.exceptions import ConfigurationError as ConfigurationError, GridGenerationError as GridGenerationError, InterpolationError as InterpolationError, ProcessingError as ProcessingError, ProjectionError as ProjectionError, RegistrationError as RegistrationError, TransformationError as TransformationError
from .grid import BaseGridGeneration as BaseGridGeneration
from .interpolation import BaseInterpolation as BaseInterpolation
from .strategy import BaseProjectionStrategy as BaseProjectionStrategy
from .transform import BaseCoordinateTransformer as BaseCoordinateTransformer

__all__ = ['BaseProjectionStrategy', 'BaseGridGeneration', 'BaseInterpolation', 'BaseCoordinateTransformer', 'ProjectionError', 'ConfigurationError', 'RegistrationError', 'ProcessingError', 'GridGenerationError', 'TransformationError', 'InterpolationError']



#!/usr/bin/env python

import os
from setuptools import setup, find_packages
import glob
import sys

# Load the README file as long_description
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

def read_requirements(path):
    with open(path) as f:
        return [
            line.strip()
            for line in f
            if line.strip() and not line.startswith("#") and not line.startswith("-e")
        ]

base_reqs = [
    "numpy>=2.3.0",
    "scipy",
    "joblib",
    "torch",
    "scikit-image",
    "opencv-python-headless",
    "pydantic>=2.0.0",
    "pyyaml",
    "open3d",
]

depth_reqs = set(read_requirements("requirements.txt"))
for req_file in glob.glob("panorai/depth/*/requirements*.txt"):
    depth_reqs.update(read_requirements(req_file))

setup(
    name="panorai",
    version="v3.0.18",
    author="Robinson Luiz Souza Garcia",
    author_email="rlsgarcia@icloud.com",
    description="Panoramic image projection and blending using Gnomonic and other spherical projections.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RobinsonGarcia/PanorAi",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Image Processing",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.7",
    install_requires=base_reqs,
    extras_require={
        "dev": [
            "pytest",
            "pytest-cov",
            "flake8",
            "black",
            "mypy",
        ],
        "docs": [
            "sphinx",
            "sphinx-rtd-theme",
        ],
        "depth": sorted(depth_reqs),
    },
    # If you have console scripts, you can uncomment and configure entry_points:
    # entry_points={
    #     "console_scripts": [
    #         "panorai-cli=panorai.cli.projection_pipeline_cli:main",
    #     ],
    # },
    # Include depth modules only when the extra is requested
    packages=[p for p in find_packages(include=['panorai', 'panorai.*'])
              if 'depth' in p and any('depth' in arg for arg in sys.argv)
              or not p.startswith('panorai.depth')],
    include_package_data=True,
    license="MIT",
    project_urls={
        "Bug Tracker": "https://github.com/RobinsonGarcia/PanorAi/issues",
        "Source Code": "https://github.com/RobinsonGarcia/PanorAi",
        "Documentation": "https://github.com/RobinsonGarcia/PanorAi/wiki",
    },
    keywords=[
        "panorama",
        "projection",
        "gnomonic",
        "spherical images",
        "3D reconstruction",
        "computer vision",
    ],
)



class PanoraiError(Exception): ...
class InvalidDataError(PanoraiError): ...
class ChannelMismatchError(PanoraiError): ...
class MetadataValidationError(PanoraiError): ...
class DataConversionError(PanoraiError): ...
class MissingChannelError(PanoraiError): ...
class FaceSetError(PanoraiError): ...
class ImageProcessingError(PanoraiError): ...
class ProjectionError(Exception): ...
class ConfigurationError(ProjectionError): ...
class RegistrationError(ProjectionError): ...
class ProcessingError(ProjectionError): ...
class GridGenerationError(ProjectionError): ...
class TransformationError(ProjectionError): ...
class InterpolationError(ProjectionError): ...



# Copyright (C) 2022-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).

"""
Script generating commandlines to generate image pairs from metadata files.
"""
import os
import glob
from tqdm import tqdm
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_dir", required=True)
    parser.add_argument("--output_dir", required=True)
    parser.add_argument("--prefix", default="", help="Commanline prefix, useful e.g. to setup environment.")
    args = parser.parse_args()

    input_metadata_filenames = glob.iglob(f"{args.input_dir}/**/metadata.json", recursive=True)

    for metadata_filename in tqdm(input_metadata_filenames):
        output_dir = os.path.join(args.output_dir, os.path.relpath(os.path.dirname(metadata_filename), args.input_dir))
        # Do not process the scene if the metadata file already exists
        if os.path.exists(os.path.join(output_dir, "metadata.json")):
            continue
        commandline = f"{args.prefix}python datasets/habitat_sim/generate_from_metadata.py --metadata_filename={metadata_filename} --output_dir={output_dir}"
        print(commandline)



from torchvision.transforms import Compose
from panorai.depth.trainers.transforms import NormalizeImage, DeterministicProjectionTransform

def build_gnomonic_projection_transform(size=518, n_angles=8):
    # Returns a Compose pipeline that expects input in dict format.
    return ComposeWithKey([
        DeterministicProjectionTransform(size=size, n_angles=n_angles),
        NormalizeImage(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])

class ComposeWithKey:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, sample_with_key):
        # For each transform, update sample_with_key["data"]
        for t in self.transforms:
            # Some transforms (like NormalizeImage) expect just the sample dict,
            # so if t has an attribute __keyless__ set to True, we pass sample_with_key["data"].
            if hasattr(t, "__keyless__") and t.__keyless__:
                sample_with_key["data"] = t(sample_with_key["data"])
            else:
                sample_with_key["data"] = t(sample_with_key)
        return sample_with_key["data"]


import numpy as np
from .equirectangular_image import EquirectangularImage as EquirectangularImage
from .gnomonic_image import GnomonicFace as GnomonicFace
from .gnomonic_imageset import GnomonicFaceSet as GnomonicFaceSet
from PIL import Image
from _typeshed import Incomplete
from typing import Literal

class DataFactory:
    data: Incomplete
    def __init__(self, data: EquirectangularImage | GnomonicFace | GnomonicFaceSet) -> None: ...
    @classmethod
    def from_array(cls, data: np.ndarray, data_type: str) -> EquirectangularImage | GnomonicFace: ...
    @classmethod
    def from_dict(cls, data: dict[str, np.ndarray], data_type: str) -> EquirectangularImage | GnomonicFace: ...
    @classmethod
    def from_list(cls, faces: list[GnomonicFace], channel_name: str = 'default') -> GnomonicFaceSet: ...
    @classmethod
    def from_pil(cls, img: Image.Image, data_type: str) -> EquirectangularImage | GnomonicFace: ...
    @classmethod
    def from_file(cls, file_path: str, data_type: Literal['equirectangular', 'gnomonic_face']) -> EquirectangularImage | GnomonicFace: ...



import numpy as np
import open3d as o3d

class PCD:
    def __init__(self, points: np.ndarray, colors: np.ndarray, radius_image: np.ndarray = None, indexes: np.ndarray = None, shape: tuple = None) -> None: ...
    def hook_bias_adjustment(self, func): ...
    @property
    def points(self) -> np.ndarray: ...
    @property
    def indexes(self) -> np.ndarray: ...
    @property
    def shape(self) -> np.ndarray: ...
    @property
    def radius_image(self) -> np.ndarray: ...
    @property
    def colors(self) -> np.ndarray: ...
    @property
    def o3d(self) -> o3d.geometry.PointCloud: ...



# panorai/projections/gnomonic_projection.py

import numpy as np
from typing import Tuple, Optional
from .registry import ProjectionRegistry
from ..utils.exceptions import ProcessingError, ConfigurationError
from .gnomonic.config import GnomonicConfig
from .gnomonic.grid import GnomonicGridGeneration
from .gnomonic.strategy import GnomonicProjectionStrategy
from .gnomonic.transform import GnomonicTransformer
from .base.interpolation import BaseInterpolation

@ProjectionRegistry.register('gnomonic')
class GnomonicProjection:
    """
    Concrete class for Gnomonic projection (NumPy-based).
    """

    def __init__(self, config: Optional[GnomonicConfig] = None,
                phi1_deg: float = 0.0, lam0_deg: float = 0.0, fov_deg: float = 90.0, **kwargs):
        try:
            # Use provided shared configuration if available
            if config is None:
                config = GnomonicConfig(phi1_deg=phi1_deg, lam0_deg=lam0_deg, fov_deg=fov_deg, **kwargs)
            self.config = config
            self.grid_generator = GnomonicGridGeneration(self.config)
            self.strategy = GnomonicProjectionStrategy(self.config)
            self.transformer = GnomonicTransformer(self.config)
            self.interpolation = BaseInterpolation(self.config)
        except Exception as e:
            raise ConfigurationError(f"Failed to initialize GnomonicProjection: {e}") from e


    def project(self, eq_img: np.ndarray) -> np.ndarray:
        """
        Project an equirectangular (NumPy) image onto the Gnomonic plane.
        """
        try:
            grid_x, grid_y = self.grid_generator.projection_grid()
            lat, lon = self.strategy.from_projection_to_spherical(grid_x, grid_y)
            map_x, map_y = self.transformer.spherical_to_image_coords(lat, lon, eq_img.shape[:2])
            projected_img = self.interpolation.interpolate(eq_img, map_x, map_y)
            return projected_img
        except Exception as e:
            raise ProcessingError(f"Error during forward projection: {e}") from e

    def back_project(self, face_img: np.ndarray, eq_shape: Tuple[int, int]) -> np.ndarray:
        """
        Back-project a Gnomonic face onto the equirectangular image plane.
        """
        if eq_shape:
            self.config.update(lat_points=eq_shape[0], lon_points=eq_shape[1])
        try:
            lon_grid, lat_grid = self.grid_generator.spherical_grid()
            x, y, mask = self.strategy.from_spherical_to_projection(lat_grid, lon_grid)
            map_x, map_y = self.transformer.projection_to_image_coords(x, y, self.config)
            eq_img = self.interpolation.interpolate(face_img, map_x, map_y, mask)
            return np.flip(eq_img, axis=0)
        except Exception as e:
            raise ProcessingError(f"Error during backward projection: {e}") from e

    # def create_face(self, eq_image: EquirectangularImage) -> GnomonicFace:
    #     """
    #     Create a GnomonicFace from an EquirectangularImage.
    #     """
    #     try:
    #         projected_channels = self.project(eq_image.data)
    #         gnomonic_face = GnomonicFace(
    #             data=projected_channels,
    #             lat=self.config["phi1_deg"],
    #             lon=self.config["lam0_deg"],
    #             fov=self.config["fov_deg"]
    #         )
    #         return gnomonic_face
    #     except Exception as e:
    #         raise ProcessingError(f"Error while creating GnomonicFace: {e}") from e


"""
Arguments:
maxR -- maximum drop radius
minR -- minimum drop radius
maxDrops -- maximum number of drops in the image
minDrops -- minimum number of drops in the image
edge_darkratio -- brightness reduction factor for drops edges
return_label -- flag defining whether a label will be returned or just an image with generated raindrops
A, B, C, D -- in this code are useless, old version is used for control bezeir
"""

cfg = {
    'maxR': 35,  # max not more then 150
    'minR': 10,
    'maxDrops': 50,
    'minDrops': 15,
    'edge_darkratio': 1.0,
    'return_label': True,
    'label_thres': 128,
    'A': (1, 4.5),
    'B': (3, 1),
    'C': (1, 3),
    'D': (3, 3)
}



# Copyright (C) 2022-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
"""
Utility script to pack metadata files of the dataset in order to be able to re-generate it elsewhere.
"""
import os
import glob
from tqdm import tqdm
import shutil
import json
from datasets.habitat_sim.paths import *
import argparse
import collections

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_dir")
    parser.add_argument("output_dir")
    args = parser.parse_args()

    input_dirname = args.input_dir
    output_dirname = args.output_dir

    input_metadata_filenames = glob.iglob(f"{input_dirname}/**/metadata.json", recursive=True)

    images_count = collections.defaultdict(lambda : 0)
    
    os.makedirs(output_dirname)
    for input_filename in tqdm(input_metadata_filenames):
        # Ignore empty files
        with open(input_filename, "r") as f:
            original_metadata = json.load(f)
            if "multiviews" not in original_metadata or len(original_metadata["multiviews"]) == 0:
                print("No views in", input_filename)
                continue

        relpath = os.path.relpath(input_filename, input_dirname)
        print(relpath)

        # Copy metadata, while replacing scene paths by generic keys depending on the dataset, for portability.
        # Data paths are sorted by decreasing length to avoid potential bugs due to paths starting by the same string pattern.
        scenes_dataset_paths = dict(sorted(SCENES_DATASET.items(), key=lambda x: len(x[1]), reverse=True))
        metadata = dict()
        for key, value in original_metadata.items():
            if key in ("scene_dataset_config_file", "scene", "navmesh") and value != "":
                known_path = False
                for dataset, dataset_path in scenes_dataset_paths.items():
                    if value.startswith(dataset_path):
                        value = os.path.join(dataset, os.path.relpath(value, dataset_path))
                        known_path = True
                        break
                if not known_path:
                    raise KeyError("Unknown path:" + value)
            metadata[key] = value

        # Compile some general statistics while packing data
        scene_split = metadata["scene"].split("/")
        upper_level = "/".join(scene_split[:2]) if scene_split[0] == "hm3d" else scene_split[0]
        images_count[upper_level] += len(metadata["multiviews"])
        
        output_filename = os.path.join(output_dirname, relpath)
        os.makedirs(os.path.dirname(output_filename), exist_ok=True)
        with open(output_filename, "w") as f:
            json.dump(metadata, f)

    # Print statistics
    print("Images count:")
    for upper_level, count in images_count.items():
        print(f"- {upper_level}: {count}")


from ..config.registry import ConfigRegistry as ConfigRegistry
from ..utils.exceptions import ConfigurationError as ConfigurationError
from _typeshed import Incomplete
from pydantic import BaseModel
from typing import Any

logger: Incomplete

class SamplerConfigModel(BaseModel):
    rotations: list[tuple] | None
    n_points: int | None
    model_config: Incomplete
    class Config:
        arbitrary_types_allowed: bool
        extra: str

class SamplerConfig:
    def __init__(self, **kwargs: Any) -> None: ...
    def update(self, **kwargs: Any) -> None: ...
    def __getattr__(self, item: str) -> Any: ...
    def __getitem__(self, key: str) -> Any: ...
    def __iter__(self): ...
